from dataclasses import dataclass
from typing import Any

from transformers import PreTrainedTokenizerBase

from turbo_alignment.dataset.pair_preferences import PairPreferenceDataCollator


@dataclass
class DDPODataCollator(PairPreferenceDataCollator):
    rm_tokenizer: PreTrainedTokenizerBase = None

    def __call__(self, examples: list[dict[str, dict[str, Any]]]) -> dict[str, Any]:
        sft_max_length, rm_max_length = 0, 0
        for ex in examples:
            sft_max_length = max(sft_max_length, len(ex['sft_inputs_w']['input_ids']))
            sft_max_length = max(sft_max_length, len(ex['sft_inputs_l']['input_ids']))
            rm_max_length = max(rm_max_length, len(ex['rm_inputs_w']['input_ids']))
            rm_max_length = max(rm_max_length, len(ex['rm_inputs_l']['input_ids']))

        batch = {
            'sft_inputs_w': dict(self._get_batch(examples, self.tokenizer, 'sft_inputs_w', sft_max_length)),
            'sft_inputs_l': dict(self._get_batch(examples, self.tokenizer, 'sft_inputs_l', sft_max_length)),
            'rm_inputs_w': dict(self._get_batch(examples, self.rm_tokenizer, 'rm_inputs_w', rm_max_length)),
            'rm_inputs_l': dict(self._get_batch(examples, self.rm_tokenizer, 'rm_inputs_l', rm_max_length)),
        }

        return batch



from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class SpecialTokensSettings(ExtraFieldsNotAllowedBaseModel):
    bos_token: str
    eos_token: str
    pad_token: str | None = None
    unk_token: str | None = None
    sep_token: str | None = None



import torch
import torch.nn.functional as F
from transformers import DataCollatorForTokenClassification


class DataCollatorWithModalityInputs(DataCollatorForTokenClassification):
    def torch_call(self, features):
        label_name = 'label' if 'label' in features[0].keys() else 'labels'
        labels = [feature[label_name] for feature in features] if label_name in features[0].keys() else None

        modality_inputs = [feature['modality_inputs'] for feature in features]

        modality_input_names = (label_name, 'modality_inputs', 'modality_tokens_mask')
        tokenizer_features = [
            {k: v for k, v in feature.items() if k not in modality_input_names} for feature in features
        ]

        batch = self.tokenizer.pad(
            tokenizer_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors='pt',
        )

        sequence_length = batch['input_ids'].shape[1]
        padding_side = self.tokenizer.padding_side

        assert padding_side == 'right'

        batch['modality_inputs'] = modality_inputs

        batch['modality_tokens_mask'] = torch.stack(
            [
                F.pad(
                    feature['modality_tokens_mask'],
                    pad=(0, sequence_length - len(feature['modality_tokens_mask'])),
                    mode='constant',
                    value=0,
                )
                for feature in features
            ]
        )

        if labels is None:
            return batch

        batch[label_name] = [
            label.tolist() + [self.label_pad_token_id] * (sequence_length - len(label)) for label in labels
        ]

        batch[label_name] = torch.tensor(batch[label_name], dtype=torch.int64)

        return batch



from typing import Callable

import torch
from torch.utils.data import Dataset
from transformers import (
    AdamW,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    TrainingArguments,
)
from transformers.data.data_collator import (
    DataCollatorForTokenClassification,
    DataCollatorMixin,
)

from turbo_alignment.cherry_picks.rag import RagCherryPickCallback
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.loaders.model.model import load_model
from turbo_alignment.constants import TRAINER_LOGS_FOLDER
from turbo_alignment.dataset.chat import InferenceChatDataset
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MetricSettingsRegistry
from turbo_alignment.modeling.rag.rag_model import RagSequenceForGeneration
from turbo_alignment.modeling.rag.rag_tokenizer import RagTokenizer
from turbo_alignment.pipelines.train.base import BaseTrainStrategy
from turbo_alignment.settings.datasets import DatasetStrategy
from turbo_alignment.settings.pipelines.train.rag import RAGTrainExperimentSettings
from turbo_alignment.trainers.multigpu import MultiGPUCherryPicksTrainer

logger = get_project_logger()


class TrainRAGStrategy(BaseTrainStrategy[RAGTrainExperimentSettings]):
    @staticmethod
    def _get_data_collator(
        experiment_settings: RAGTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> Callable:
        return DataCollatorForTokenClassification(tokenizer, pad_to_multiple_of=8)

    @staticmethod
    def _get_cherry_pick_callback(
        experiment_settings: RAGTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> RagCherryPickCallback:
        cherry_pick_settings = experiment_settings.cherry_pick_settings

        cherry_pick_datasets = DatasetLoader[InferenceChatDataset](InferenceChatDataset).load_datasets(
            cherry_pick_settings.dataset_settings, tokenizer=tokenizer, strategy=DatasetStrategy.INFERENCE
        )

        metrics = [
            Metric.by_name(metric.type)(MetricSettingsRegistry.by_name(metric.type)(**metric.parameters))
            for metric in cherry_pick_settings.metric_settings
        ]

        return RagCherryPickCallback(
            cherry_pick_settings=cherry_pick_settings,
            datasets=cherry_pick_datasets,
            metrics=metrics,
        )

    @staticmethod
    def _get_additional_special_tokens(
        experiment_settings: RAGTrainExperimentSettings,  # type: ignore[override]
    ) -> list[str]:
        gen_settings = experiment_settings.model_settings.generator_settings
        embeddings_initialization_strategy = gen_settings.embeddings_initialization_strategy
        return list(embeddings_initialization_strategy.keys()) if embeddings_initialization_strategy else []

    @staticmethod
    def _get_training_args(experiment_settings: RAGTrainExperimentSettings) -> TrainingArguments:
        return TrainingArguments(
            output_dir=str(experiment_settings.log_path / TRAINER_LOGS_FOLDER),
            **experiment_settings.trainer_settings.dict(),
        )

    @staticmethod
    def _get_trainer(
        training_args: TrainingArguments,
        experiment_settings: RAGTrainExperimentSettings,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: DataCollatorMixin,
        **_kwargs,
    ) -> MultiGPUCherryPicksTrainer:
        trainer = MultiGPUCherryPicksTrainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            callbacks=[],
            data_collator=data_collator,
            tokenizer=tokenizer,
        )
        generator_parameters = model.rag.generator.parameters()
        question_encoder_parameters = model.rag.question_encoder.parameters()
        encoder_lr = experiment_settings.model_settings.retrieval_settings.encoder_learning_rate
        optimizer = AdamW(
            [
                {'params': generator_parameters, 'lr': training_args.learning_rate},
                {
                    'params': question_encoder_parameters,
                    'lr': encoder_lr if encoder_lr is not None else training_args.learning_rate,
                },
            ]
        )
        trainer.optimizer = optimizer
        return trainer

    @staticmethod
    def _load_tokenizer(experiment_settings: RAGTrainExperimentSettings) -> RagTokenizer:
        tokenizer = RagTokenizer(
            model_settings=experiment_settings.model_settings,
            tokenizer_path=experiment_settings.tokenizer_settings.tokenizer_path,
            use_fast=experiment_settings.tokenizer_settings.use_fast,
        )
        return tokenizer

    @staticmethod
    def _load_model(
        experiment_settings: RAGTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
    ) -> torch.nn.Module | PreTrainedModel:
        model_settings = experiment_settings.model_settings

        generator = load_model(model_settings.generator_settings, tokenizer.generator)
        question_encoder = load_model(model_settings.question_encoder_settings, tokenizer.question_encoder)

        model = RagSequenceForGeneration(model_settings, generator, question_encoder, tokenizer)
        return model

    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: DataCollatorMixin) -> None:
        logger.info(f'Train sample example:\n{dataset[0]}')
        logger.info('Example text check: {example}'.format(example=self.tokenizer.decode(dataset[0]['input_ids'])))
        logger.info(
            'Input ids check: {input_ids}'.format(input_ids=collator([dataset[0], dataset[1]])['input_ids'][0])
        )
        logger.info('Mask check: {mask}'.format(mask=collator([dataset[0], dataset[1]])['attention_mask'][0]))
        logger.info('Labels check: {labels}'.format(labels=collator([dataset[0], dataset[1]])['labels'][0]))



import random

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.dataset.sampling.models import SamplingDatasetRecord
from turbo_alignment.pipelines.sampling.base import BaseSamplingStrategy
from turbo_alignment.settings.pipelines.sampling import RandomSamplingSettings


class RandomSamplingStrategy(BaseSamplingStrategy):
    def sample(self, experiment_settings: RandomSamplingSettings) -> list[SamplingDatasetRecord]:
        if experiment_settings.dataset_settings.sources[0].records_path:
            records = [
                SamplingDatasetRecord(**record)
                for record in read_jsonl(experiment_settings.dataset_settings.sources[0].records_path)
            ]
        else:
            raise ValueError('records_path should be not None!')

        for record in records:
            record.answers = random.sample(record.answers, min(experiment_settings.N, len(record.answers)))

        return records



from typing import Sequence

from turbo_alignment.settings.metric import MetricSettings
from turbo_alignment.settings.modality import (
    Modality,
    ModalityEncoderSettings,
    ModalityProjectorType,
)
from turbo_alignment.settings.model import PreTrainedMultiModalModel
from turbo_alignment.settings.pipelines.inference.base import (
    InferenceExperimentSettings,
)
from turbo_alignment.settings.pipelines.inference.chat import (
    ChatSingleModelInferenceSettings,
)


class MultimodalSingleModelInferenceSettings(ChatSingleModelInferenceSettings):
    modality_encoder_settings_mapping: dict[Modality, ModalityEncoderSettings | None]
    modality_projector_mapping: dict[Modality, ModalityProjectorType | None]
    model_settings: PreTrainedMultiModalModel
    metric_settings: list[MetricSettings] | None


class MultimodalInferenceExperimentSettings(InferenceExperimentSettings):
    inference_settings: Sequence[MultimodalSingleModelInferenceSettings]



import torch

from turbo_alignment.modeling.multimodal.encoders import ModalityEncoderRegistry
from turbo_alignment.modeling.multimodal.encoders.base import BaseModalityEncoder
from turbo_alignment.settings.modality import Modality, ModalityEncoderSettings


class MultimodalMixin:
    @staticmethod
    def _load_modality_encoders(
        modality_encoder_settings_mapping: dict[Modality, ModalityEncoderSettings | None],
        device: torch.device,
        dtype: torch.dtype,
    ) -> dict[Modality, BaseModalityEncoder]:
        encoders_dict: dict[Modality, BaseModalityEncoder] = {}

        for modality, encoder_settings in modality_encoder_settings_mapping.items():
            if encoder_settings:
                encoder = ModalityEncoderRegistry.by_name(encoder_settings.modality_encoder_type)(
                    encoder_path=encoder_settings.encoder_path,
                    is_pickle=encoder_settings.is_pickle,
                )
                encoder = encoder.to(device)
                encoder = encoder.to(dtype)
                encoders_dict[modality] = encoder

        return encoders_dict



from pydantic_settings import BaseSettings


class ClearMLSettings(BaseSettings):
    project_name: str
    task_name: str
    tags: list[str] = []

    __name__ = 'ClearMLSettings'

    class Config:
        env_prefix: str = 'CLEARML_'



from .base import DatasetStrategy, DatasetType
from .chat import ChatMultiDatasetSettings
from .classification import ClassificationMultiDatasetSettings
from .pair_preference import PairPreferenceMultiDatasetSettings



from pathlib import Path

import typer

from turbo_alignment import pipelines
from turbo_alignment.cli.app import app
from turbo_alignment.settings import pipelines as pipeline_settings


@app.command(name='train_sft', help='Run PEFT pipeline')
def train_sft_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ...,
        '--experiment_settings_path',
        exists=True,
        help='Path to experiment config file',
    )
) -> None:
    experiment_settings = pipeline_settings.SftTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainSFTStrategy().run(experiment_settings)


@app.command(name='train_dpo', help='Run DPO pipeline')
def train_dpo_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ...,
        '--experiment_settings_path',
        exists=True,
        help='Path to experiment config file',
    )
) -> None:
    experiment_settings = pipeline_settings.DPOTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainDPOStrategy().run(experiment_settings)


@app.command(name='train_kto', help='Run KTO pipeline')
def train_kto_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ...,
        '--experiment_settings_path',
        exists=True,
        help='Path to experiment config file',
    )
) -> None:
    experiment_settings = pipeline_settings.KTOTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainKTOStrategy().run(experiment_settings)


@app.command(name='train_ddpo', help='Run DDPO pipeline')
def train_ddpo_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ...,
        '--experiment_settings_path',
        exists=True,
        help='Path to experiment config file',
    )
) -> None:
    experiment_settings = pipeline_settings.DDPOTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainDDPOStrategy().run(experiment_settings)


@app.command(name='train_rm', help='Run RM pipeline')
def train_rm_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ...,
        '--experiment_settings_path',
        exists=True,
        help='Path to experiment config file',
    )
) -> None:
    experiment_settings = pipeline_settings.RMTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainRMStrategy().run(experiment_settings)


@app.command(name='train_classification', help='Train Classifier')
def classification_training(
    experiment_settings_path: Path = typer.Option(
        ...,
        '--experiment_settings_path',
        exists=True,
        help='Path to experiment config file',
    )
) -> None:
    experiment_settings = pipeline_settings.ClassificationTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainClassificationStrategy().run(experiment_settings)


@app.command(name='train_multimodal', help='Train Multimodal')
def multimodal_training(
    experiment_settings_path: Path = typer.Option(
        ..., '--experiment_settings_path', exists=True, help='Path to experiment config file'
    )
) -> None:
    experiment_settings = pipeline_settings.MultimodalTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainMultimodalStrategy().run(experiment_settings)


@app.command(name='train_rag', help='Train RAG')
def rag_training(
    experiment_settings_path: Path = typer.Option(
        ..., '--experiment_settings_path', exists=True, help='Path to experiment config file'
    )
) -> None:
    experiment_settings = pipeline_settings.RAGTrainExperimentSettings.parse_file(experiment_settings_path)
    pipelines.TrainRAGStrategy().run(experiment_settings)



from typing import Literal

from turbo_alignment.settings.datasets.base import (
    BaseDatasetSettings,
    DatasetType,
    MultiDatasetSettings,
)
from turbo_alignment.settings.datasets.chat import ChatDatasetSettings


class KTODatasetSettings(BaseDatasetSettings):
    dataset_type: Literal[DatasetType.KTO] = DatasetType.KTO
    chat_settings: ChatDatasetSettings


class KTOMultiDatasetSettings(KTODatasetSettings, MultiDatasetSettings):
    ...



from typing import Literal

from peft import PeftType, TaskType

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class BasePeftSettings(ExtraFieldsNotAllowedBaseModel):
    name: PeftType
    task_type: TaskType = TaskType.CAUSAL_LM


class LoraSettings(BasePeftSettings):
    name: Literal[PeftType.LORA] = PeftType.LORA
    r: int = 16
    lora_alpha: int = 16
    lora_dropout: float = 0.05
    bias: str = 'none'
    target_modules: list[str] = ['q_proj', 'v_proj']
    modules_to_save: list[str] | None = None


class PrefixTuningSettings(BasePeftSettings):
    name: Literal[PeftType.PREFIX_TUNING] = PeftType.PREFIX_TUNING
    encoder_hidden_size: int
    prefix_projection: bool


class PromptTuningSettings(BasePeftSettings):
    name: Literal[PeftType.PROMPT_TUNING] = PeftType.PROMPT_TUNING
    num_virtual_tokens: int = 32
    prompt_tuning_init_text: str | None = None


class PTuningSettings(BasePeftSettings):
    name: Literal[PeftType.P_TUNING] = PeftType.P_TUNING
    num_virtual_tokens: int = 32
    encoder_reparameterization_type: str = 'MLP'


PEFT_TYPE = PrefixTuningSettings | LoraSettings | PromptTuningSettings | PTuningSettings



from turbo_alignment.metrics.distinctness import DistinctnessMetric
from turbo_alignment.metrics.diversity import DiversityMetric
from turbo_alignment.metrics.kl import KLMetric
from turbo_alignment.metrics.length import LengthMetric
from turbo_alignment.metrics.meteor import MeteorMetric
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.perplexity import PerplexityMetric
from turbo_alignment.metrics.registry import *
from turbo_alignment.metrics.retrieval_utility import RetrievalUtilityMetric
from turbo_alignment.metrics.reward import RewardMetric
from turbo_alignment.metrics.rouge import RougeMetric
from turbo_alignment.metrics.self_bleu import SelfBleuMetric



from .classification import ClassificationTrainer
from .custom_loss import CustomLossTrainer
from .ddpo import DDPOTrainer
from .dpo import DPOTrainer
from .kto import KTOTrainer
from .multigpu import MultiGPUCherryPicksTrainer
from .multimodal import MultimodalTrainer
from .rm import RMTrainer



from typing import Callable

from torch import nn
from torch.utils.data import Dataset
from transformers import (
    DefaultFlowCallback,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    PrinterCallback,
    ProgressCallback,
    Trainer,
    TrainerCallback,
    TrainerControl,
    TrainingArguments,
)
from transformers.integrations import get_reporting_integration_callbacks

from turbo_alignment.common.tf.callbacks.common import MetricsCallbackHandler


class MultiGPUCherryPicksTrainer(Trainer):
    def __init__(
        self,
        model: PreTrainedModel | nn.Module,
        data_collator: Callable,
        args: TrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        tokenizer: PreTrainedTokenizerBase | None = None,
        callbacks: list[TrainerCallback] | None = None,
        **kwargs,
    ):
        ref_model = kwargs.pop('ref_model', None)
        metrics_kwargs = kwargs.pop('metrics_kwargs', None)

        super().__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            callbacks=callbacks,
            **kwargs,
        )

        default_callbacks = [DefaultFlowCallback] + get_reporting_integration_callbacks(self.args.report_to)
        callbacks = default_callbacks if callbacks is None else default_callbacks + callbacks
        self.callback_handler = MetricsCallbackHandler(
            callbacks,
            model,
            tokenizer,
            None,
            None,
            ref_model=ref_model,
            accelerator=self.accelerator,
            metrics_kwargs=metrics_kwargs,
        )
        self.add_callback(PrinterCallback if self.args.disable_tqdm else ProgressCallback)
        self.control: TrainerControl = self.callback_handler.on_init_end(self.args, self.state, self.control)



import numpy as np
from nltk.translate.bleu_score import SmoothingFunction, sentence_bleu

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import SelfBleuSettings
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.SELF_BLEU)
class SelfBleuMetric(Metric):
    def __init__(self, settings: SelfBleuSettings) -> None:
        super().__init__(settings=settings)
        self._settings: SelfBleuSettings = settings

    def compute(self, **kwargs) -> list[MetricResults]:
        predictions: list[list[str]] = kwargs.get('predictions', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if predictions is None:
            raise ValueError('predictions should not be None')

        return [
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(
                        label=dataset_name + '@@' + 'self_bleu',
                        values=[self.self_bleu(answers) for answers in predictions],
                    )
                ],
                need_average=need_average,
            )
            for need_average in self._settings.need_average
        ]

    def self_bleu(self, answers: list[str]) -> float:
        weight = tuple((1.0 / self._settings.ngram for _ in range(self._settings.ngram)))
        result = []
        sentence_num = len(answers)
        for index in range(sentence_num):
            hypothesis = answers[index]
            other = answers[:index] + answers[index + 1 :]
            result.append(self._calc_bleu(other, hypothesis, weight))
        return np.mean(result).item()

    def _calc_bleu(self, reference: list[str], hypothesis: str, weight: tuple[float, ...]) -> list[float]:
        return sentence_bleu(reference, hypothesis, weight, smoothing_function=SmoothingFunction().method1)



from pathlib import Path

import cv2
import torch
from transformers import CLIPProcessor

from turbo_alignment.common.data.multimodal.image.base import BaseImageReader
from turbo_alignment.common.data.multimodal.registry import ImageModalityReaderRegistry
from turbo_alignment.settings.modality import ModalityReader


@ImageModalityReaderRegistry.register(ModalityReader.CLIP)
class CLIPImageReader(BaseImageReader):
    def __init__(self, reader_path: Path, processor_clip: CLIPProcessor | None = None):
        if processor_clip:
            self.processor_clip = processor_clip
        else:
            self.processor_clip = CLIPProcessor.from_pretrained(reader_path)

    def read(self, path: str) -> torch.Tensor:
        image = cv2.imread(path)
        if image is None:
            raise OSError(f'Image not found: {path}')

        return self.processor_clip(images=image, return_tensors='pt')['pixel_values']



# mypy: ignore-errors
import logging
from typing import Any

logger = logging.getLogger(__name__)


class ConfigurationError(Exception):
    def __reduce__(self) -> str | tuple[Any, ...]:
        return type(self), (self.message,)

    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message



from typing import Annotated, Literal

from pydantic import BaseModel, Field

from turbo_alignment.dataset.base.models import DatasetRecord
from turbo_alignment.dataset.chat.models import ChatMessageRole
from turbo_alignment.settings.modality import Modality


class MultimodalChatMessage(BaseModel):
    role: ChatMessageRole
    type: Modality
    disable_loss: bool = False


class MultimodalTextMessage(MultimodalChatMessage):
    content: str
    type: Literal[Modality.TEXT] = Modality.TEXT


class MultimodalFileMessage(MultimodalChatMessage):
    content: str


class MultimodalImageMessage(MultimodalFileMessage):
    type: Literal[Modality.IMAGE] = Modality.IMAGE


class MultimodalAudioMessage(MultimodalFileMessage):
    type: Literal[Modality.AUDIO] = Modality.AUDIO


class MultimodalDatasetRecord(DatasetRecord):
    messages: list[
        Annotated[MultimodalImageMessage | MultimodalAudioMessage | MultimodalTextMessage, Field(discriminator='type')]
    ]
    label: str | None = None



from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class GeneratorTransformersSettings(ExtraFieldsNotAllowedBaseModel):
    num_beams: int = 1
    max_new_tokens: int = 15
    repetition_penalty: float = 1.0
    num_return_sequences: int = 1
    do_sample: bool = True
    top_p: float = 1.0
    top_k: int = 50
    temperature: float = 1.0
    stop_strings: str | list[str] = '</s>'



# mypy: ignore-errors
import logging
import os
from pathlib import Path

import cached_path as _cached_path

logger = logging.getLogger(__name__)


def cached_path(
    url_or_filename: str | os.PathLike,
    cache_dir: str | Path = None,
    extract_archive: bool = False,
    force_extract: bool = False,
) -> str:
    return str(
        _cached_path.cached_path(
            url_or_filename,
            cache_dir=cache_dir,
            extract_archive=extract_archive,
            force_extract=force_extract,
        )
    )



from pathlib import Path

import typer

from turbo_alignment import pipelines
from turbo_alignment.cli.app import app
from turbo_alignment.common.tf.merge_adapters_to_base import peft_to_base_model
from turbo_alignment.settings.datasets.multimodal import (
    MultimodalDatasetProcessingSettings,
)
from turbo_alignment.settings.pipelines import MergeAdaptersToBaseModelSettings


@app.command(
    name='preprocess_multimodal_dataset', help='Read and encode multimodal objects from image / audio dataset'
)
def process_multimodal_dataset(
    settings_path: Path = typer.Option(
        ..., '--settings_path', exists=True, help='Path to multimodal dataset processing settings'
    )
):
    settings = MultimodalDatasetProcessingSettings.parse_file(settings_path)
    pipelines.PreprocessMultimodalDatasetStrategy().run(settings)


@app.command(name='merge_adapters_to_base', help='Merge peft adapters into base model')
def merge_adapters_to_base_entrypoint(
    settings_path: Path = typer.Option(..., '--settings_path', exists=True, help='Path to script config file')
) -> None:
    settings = MergeAdaptersToBaseModelSettings.parse_file(settings_path)
    peft_to_base_model(settings)



import os
from pathlib import Path

from transformers import AutoTokenizer

from turbo_alignment.settings.rag_model import RAGPreTrainedModelSettings


class RagTokenizer:
    def __init__(self, model_settings: RAGPreTrainedModelSettings, tokenizer_path: str | None, use_fast: bool = False):
        qe_config = model_settings.question_encoder_settings
        gen_config = model_settings.generator_settings

        if tokenizer_path is not None:
            question_encoder_path = Path(tokenizer_path) / Path('question_encoder_tokenizer')
            generator_path = Path(tokenizer_path) / Path('generator_tokenizer')
            self.question_encoder = AutoTokenizer.from_pretrained(question_encoder_path, use_fast=use_fast)
            self.generator = AutoTokenizer.from_pretrained(generator_path, use_fast=use_fast)
        else:
            self.question_encoder = AutoTokenizer.from_pretrained(qe_config.model_path, use_fast=use_fast)
            self.generator = AutoTokenizer.from_pretrained(gen_config.model_path, use_fast=use_fast)

        self.current_tokenizer = self.generator

        self.bos_token_id = self.generator.bos_token_id
        self.eos_token_id = self.generator.eos_token_id
        self.pad_token_id = self.generator.pad_token_id
        self.unk_token_id = self.generator.unk_token_id
        self.sep_token_id = self.generator.sep_token_id

        self.padding_side = self.generator.padding_side

    def save_pretrained(self, save_directory) -> None:
        if os.path.isfile(save_directory):
            raise ValueError(f'Provided path ({save_directory}) should be a directory, not a file')
        os.makedirs(save_directory, exist_ok=True)
        question_encoder_path = os.path.join(save_directory, 'question_encoder_tokenizer')
        generator_path = os.path.join(save_directory, 'generator_tokenizer')
        self.question_encoder.save_pretrained(question_encoder_path)
        self.generator.save_pretrained(generator_path)

    def __call__(self, *args, **kwargs):
        return self.current_tokenizer(*args, **kwargs)

    def __len__(self):
        return len(self.generator)

    def batch_decode(self, *args, **kwargs) -> list[str]:
        return self.generator.batch_decode(*args, **kwargs)

    def decode(self, *args, **kwargs) -> str:
        return self.generator.decode(*args, **kwargs)

    def encode(self, *args, **kwargs) -> list[int]:
        return self.generator.encode(*args, **kwargs)

    def _switch_to_input_mode(self) -> None:
        self.current_tokenizer = self.question_encoder

    def _switch_to_target_mode(self) -> None:
        self.current_tokenizer = self.generator

    def pad(self, *args, **kwargs):
        return self.generator.pad(*args, **kwargs)

    def add_special_tokens(self, special_tokens_dict) -> int:
        added_tokens = self.generator.add_special_tokens(special_tokens_dict)
        # TODO: fix this
        if list(special_tokens_dict.keys())[0] == 'sep_token':
            self.sep_token_id = self.generator.sep_token_id
        if list(special_tokens_dict.keys())[0] == 'pad_token':
            self.pad_token_id = self.generator.pad_token_id
        return added_tokens

    def get_added_vocab(self) -> dict[str, int]:
        return self.generator.get_added_vocab()



import torch
from transformers import WhisperForConditionalGeneration

from turbo_alignment.modeling.multimodal.encoders.audio.base import BaseAudioEncoder
from turbo_alignment.modeling.multimodal.encoders.registry import (
    ModalityEncoderRegistry,
)
from turbo_alignment.settings.modality import ModalityEncoderType


@ModalityEncoderRegistry.register(ModalityEncoderType.WHISPER)
class WhisperAudioModeling(BaseAudioEncoder):
    def __init__(self, model_whisper: WhisperForConditionalGeneration | None = None):
        super().__init__()

        if model_whisper is not None:
            self.model_whisper = model_whisper
        else:
            self.model_whisper = WhisperForConditionalGeneration.from_pretrained('openai/whisper-large-v3')
        self.whisper_encoder = self.model_whisper.model.encoder

    def encode(self, inputs: torch.Tensor) -> torch.Tensor:
        audio_embeddings = self.whisper_encoder(inputs).last_hidden_state
        return audio_embeddings.mean(dim=1)

    @property
    def emb_dim(self) -> int:
        return self.whisper_encoder.config.d_model



from pathlib import Path

import soundfile as sf
import torch
from transformers import AutoFeatureExtractor

from turbo_alignment.common.data.multimodal.image.base import BaseImageReader
from turbo_alignment.common.data.multimodal.registry import AudioModalityReaderRegistry
from turbo_alignment.settings.modality import ModalityReader


@AudioModalityReaderRegistry.register(ModalityReader.WHISPER)
class WhisperAudioReader(BaseImageReader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.whisper_feature_extractor = AutoFeatureExtractor.from_pretrained('openai/whisper-large-v3')

    def read(self, path: str) -> torch.Tensor:
        data, _ = sf.read(path)
        inputs = self.whisper_feature_extractor(
            data, sampling_rate=self.whisper_feature_extractor.sampling_rate, return_tensors='pt'
        )
        return inputs['input_features'][0]



from abc import ABC
from pathlib import Path

import torch
from torch.nn import Module
from transformers import PreTrainedModel

from turbo_alignment.modeling.multimodal.encoders.base import BaseModalityEncoder
from turbo_alignment.settings.modality import Modality, ModalityProjectorType


class BaseMultiModalModeling(Module, ABC):
    def __init__(
        self,
        language_model: PreTrainedModel,
        encoders: dict[Modality, BaseModalityEncoder],
        n_modality_embs: int,
        modality_projector_mapping: dict[Modality, ModalityProjectorType | None],
        modality_projector_initialization_mapping: dict[Modality, Path | None] | None,
        peft: bool = True,  # TODO: Pre-training without PEFT adapters
    ) -> None:
        """
        :param n_modality_embs: The number of projections. If 1, the whole\
            image will be projected into one embedding of lm
        :param peft: If True, assume that lm was initialized with peft
        """
        super().__init__()

        self.language_model = language_model
        self.n_modality_embs = n_modality_embs
        self.modality_projector_mapping = modality_projector_mapping
        self.modality_projector_initialization_mapping = modality_projector_initialization_mapping
        self.peft = peft

        self.language_model_dim = (
            language_model.base_model.model.model.embed_tokens.modules_to_save.default.weight.shape[1]
        )
        self.encoders = torch.nn.ModuleDict(encoders)  # type: ignore[arg-type]

        for encoder in self.encoders.values():
            encoder.eval()

            for param in encoder.parameters():
                param.requires_grad = False



from typing import Callable

from torch.utils.data import Dataset
from transformers import PreTrainedModel, PreTrainedTokenizerBase
from transformers.data.data_collator import DataCollatorMixin

from turbo_alignment.cherry_picks.chat import ChatCherryPickCallback
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.loaders.model.model import load_model
from turbo_alignment.constants import TRAINER_LOGS_FOLDER
from turbo_alignment.dataset.chat.chat import InferenceChatDataset
from turbo_alignment.dataset.kto.collators import KTODataCollator
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MetricSettingsRegistry
from turbo_alignment.pipelines.train.base import BaseTrainStrategy
from turbo_alignment.settings.datasets.base import DatasetStrategy
from turbo_alignment.settings.pipelines.train.kto import KTOTrainExperimentSettings
from turbo_alignment.trainers.kto import KTOTrainer, KTOTrainingArguments

logger = get_project_logger()


class TrainKTOStrategy(BaseTrainStrategy[KTOTrainExperimentSettings]):
    @staticmethod
    def _get_data_collator(
        experiment_settings: KTOTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> Callable:
        return KTODataCollator(tokenizer=tokenizer, add_labels=True)

    @staticmethod
    def _get_cherry_pick_callback(
        experiment_settings: KTOTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> ChatCherryPickCallback:
        cherry_pick_settings = experiment_settings.cherry_pick_settings

        cherry_pick_datasets = DatasetLoader[InferenceChatDataset](InferenceChatDataset).load_datasets(
            cherry_pick_settings.dataset_settings, tokenizer=tokenizer, strategy=DatasetStrategy.INFERENCE
        )

        metrics = [
            Metric.by_name(metric.type)(MetricSettingsRegistry.by_name(metric.type)(**metric.parameters))
            for metric in cherry_pick_settings.metric_settings
        ]

        return ChatCherryPickCallback(
            cherry_pick_settings=cherry_pick_settings,
            datasets=cherry_pick_datasets,
            metrics=metrics,
        )

    @staticmethod
    def _get_training_args(experiment_settings: KTOTrainExperimentSettings) -> KTOTrainingArguments:
        return KTOTrainingArguments(
            output_dir=str(experiment_settings.log_path / TRAINER_LOGS_FOLDER),
            label_names=[],
            remove_unused_columns=False,
            **experiment_settings.trainer_settings.dict(),
        )

    @staticmethod
    def _get_trainer(
        training_args: KTOTrainingArguments,
        experiment_settings: KTOTrainExperimentSettings,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: Callable,
    ):
        model.config.use_cache = not experiment_settings.trainer_settings.gradient_checkpointing

        extra_args = {}
        if experiment_settings.trainer_settings.use_ref_model:
            ref_model = load_model(experiment_settings.model_settings, tokenizer)
            for _, param in ref_model.named_parameters():
                param.requires_grad = False

            extra_args = {'ref_model': ref_model}

        return KTOTrainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            callbacks=[],
            data_collator=data_collator,
            tokenizer=tokenizer,
            **extra_args,
        )

    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: DataCollatorMixin) -> None:
        logger.info(f'Train sample example:\n{dataset[0]}')



from abc import ABC
from pathlib import Path
from typing import Any, overload

from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.base.base import AlignmentDataset
from turbo_alignment.dataset.chat.chat import TrainChatDataset
from turbo_alignment.dataset.chat.models import ChatDatasetRecord
from turbo_alignment.dataset.classification.models import ClassificationDatasetRecord
from turbo_alignment.dataset.registry import ClassificationDatasetTypeRegistry
from turbo_alignment.settings.datasets.base import (
    DatasetSourceSettings,
    DatasetStrategy,
)
from turbo_alignment.settings.datasets.classification import (
    ClassificationDatasetSettings,
)

logger = get_project_logger()


class ClassificationDataset(AlignmentDataset[ClassificationDatasetRecord], ABC):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: ClassificationDatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
    ):
        settings.chat_settings.only_answer_loss = False
        self._chat_dataset = TrainChatDataset(
            source=source,
            settings=settings.chat_settings,
            tokenizer=tokenizer,
            read=False,
        )
        super().__init__(source=source, settings=settings, tokenizer=tokenizer)
        self.settings: ClassificationDatasetSettings = settings

        self._read()

    def _encode(self, records: list[ClassificationDatasetRecord], inference: bool) -> list[dict[str, Any] | None]:
        chat_records: list[ChatDatasetRecord] = [
            ChatDatasetRecord(id=record.id, messages=record.messages) for record in records
        ]

        tokenized_chat_records = self._chat_dataset.convert_records(chat_records)

        output: list[dict[str, Any] | None] = []
        for record, tokenized_record in zip(records, tokenized_chat_records):
            if not tokenized_record:
                continue

            classification_tokens = {k: v.squeeze(0) for k, v in tokenized_record.items() if k != 'labels'}
            classification_tokens['labels'] = record.label

            if inference:
                classification_tokens['id'] = record.id

            output.append(classification_tokens)

        return output

    @staticmethod
    @overload
    def _read_records(records: Path) -> list[ClassificationDatasetRecord]:
        ...

    @staticmethod
    @overload
    def _read_records(records: list[dict]) -> list[ClassificationDatasetRecord]:
        ...

    @staticmethod
    def _read_records(records) -> list[ClassificationDatasetRecord]:
        if isinstance(records, Path):
            return [ClassificationDatasetRecord(**record) for record in read_jsonl(records)]
        if isinstance(records, list):
            return [ClassificationDatasetRecord(**record) for record in records]
        raise NotImplementedError


@ClassificationDatasetTypeRegistry.register(DatasetStrategy.TRAIN)
class TrainClassificationDataset(ClassificationDataset):
    def convert_records(self, records: list[ClassificationDatasetRecord]) -> list[dict[str, Any] | None]:
        return self._encode(records, inference=False)


@ClassificationDatasetTypeRegistry.register(DatasetStrategy.INFERENCE)
class InferenceClassificationDataset(ClassificationDataset):
    def convert_records(self, records: list[ClassificationDatasetRecord]) -> list[dict[str, Any] | None]:
        return self._encode(records, inference=True)



import turbo_alignment.dataset.chat
import turbo_alignment.dataset.classification
import turbo_alignment.dataset.ddpo
import turbo_alignment.dataset.kto
import turbo_alignment.dataset.multimodal
import turbo_alignment.dataset.pair_preferences
import turbo_alignment.dataset.sampling



from enum import Enum

from pydantic_settings import BaseSettings


class WandbMode(str, Enum):
    ONLINE: str = 'online'
    OFFLINE: str = 'offline'
    DISABLED: str = 'disabled'


class WandbSettings(BaseSettings):
    project_name: str
    run_name: str
    entity: str
    tags: list[str] = []
    notes: str | None = None
    mode: WandbMode = WandbMode.ONLINE

    __name__ = 'WandbSettings'

    class Config:
        env_prefix: str = 'WANDB_'



from typing import Iterable

import torch
from PIL import Image
from transformers import PreTrainedTokenizerBase

import wandb
from turbo_alignment.cherry_picks.base import CherryPickCallbackBase
from turbo_alignment.dataset.multimodal import InferenceMultimodalDataset
from turbo_alignment.generators.multimodal import MultimodalGenerator
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.settings.cherry_pick import MultimodalCherryPickSettings
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults


class MultimodalCherryPickCallback(CherryPickCallbackBase[InferenceMultimodalDataset]):
    def __init__(
        self,
        cherry_pick_settings: MultimodalCherryPickSettings,
        datasets: Iterable[InferenceMultimodalDataset],
        metrics: list[Metric],
    ) -> None:
        super().__init__(cherry_pick_settings=cherry_pick_settings, datasets=datasets, metrics=metrics)
        self._custom_generation_settings = cherry_pick_settings.custom_generation_settings
        self._generator_transformers_settings = cherry_pick_settings.generator_transformers_settings

    def _get_dataset_metrics(
        self,
        dataset: InferenceMultimodalDataset,
        model: torch.nn.Module,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> list[MetricResults]:
        generator = MultimodalGenerator(
            model=model,
            tokenizer=tokenizer,
            transformers_settings=self._generator_transformers_settings,
            custom_generation_settings=self._custom_generation_settings,
        )

        generations = generator.generate_from_dataset(dataset)

        metrics_kwargs: dict = kwargs.get('metrics_kwargs', {})

        prompts = [dataset[i]['prompt'] for i in range(len(dataset))]
        string_answers = [[answer.content for answer in g.answers] for g in generations]
        string_labels = [[g.messages[-1].content] * len(g.answers) for g in generations]

        modality_object_paths_batch = [dataset[i]['modality_object_paths'] for i in range(len(dataset))]

        wandb_images: list[wandb.Image | None] = []
        wandb_audios: list[wandb.Audio | None] = []
        wandb_paths: list[str | None] = []

        for path_batch in modality_object_paths_batch:
            if len(path_batch) > 0:
                path = path_batch[0]  # TODO: How to display multiple images in wandb?
                if path.split('.')[-1] == 'jpg':
                    try:
                        with Image.open(path) as img:
                            wandb_images.append(wandb.Image(img, caption=f'Path {path}'))
                    except FileNotFoundError:
                        wandb_images.append(None)
                    wandb_audios.append(None)
                elif path.split('.')[-1] in ('wav', 'flac'):
                    wandb_audios.append(wandb.Audio(path, caption=f'Path {path}'))
                    wandb_images.append(None)
                else:
                    wandb_images.append(None)
                    wandb_audios.append(None)
                wandb_paths.append(path)
            else:
                wandb_paths.append(None)
                wandb_images.append(None)
                wandb_audios.append(None)

        metric_outputs = [
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'image', values=wandb_images)
                ]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'audio', values=wandb_audios)
                ]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'modality_object_path', values=wandb_paths)
                ]
            ),
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'prompt', values=prompts)]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'labels', values=string_labels)
                ]
            ),
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset.source.name + '@@' + 'answer', values=string_answers)
                ]
            ),
        ]

        for metric in self._metrics:
            metric_results = metric.compute(
                model=model,
                dataset=dataset,
                references=string_labels,
                predictions=string_answers,
                tokenizer=tokenizer,
                dataset_name=dataset.source.name,
                metrics_kwargs=metrics_kwargs,
            )

            metric_outputs.extend(metric_results)

        return metric_outputs



from typing import Callable

from torch.utils.data import Dataset
from transformers import PreTrainedModel, PreTrainedTokenizerBase, TrainingArguments
from transformers.data.data_collator import DataCollatorMixin

from turbo_alignment.cherry_picks.rm import RmCherryPickCallback
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.constants import TRAINER_LOGS_FOLDER
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.dataset.pair_preferences import (
    PairPreferenceDataCollator,
    PairPreferenceDataset,
)
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MetricSettingsRegistry
from turbo_alignment.metrics.reward import compute_metrics as compute_rm_metrics
from turbo_alignment.pipelines.train.base import BaseTrainStrategy
from turbo_alignment.settings.datasets import DatasetStrategy
from turbo_alignment.settings.pipelines import RMTrainExperimentSettings
from turbo_alignment.trainers.rm import RMTrainer

logger = get_project_logger()


class TrainRMStrategy(BaseTrainStrategy[RMTrainExperimentSettings]):
    @staticmethod
    def _get_data_collator(
        experiment_settings: RMTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **_kwargs,
    ) -> Callable:
        return PairPreferenceDataCollator(tokenizer=tokenizer, add_labels=False)

    @staticmethod
    def _get_cherry_pick_callback(
        experiment_settings: RMTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **_kwargs,
    ) -> RmCherryPickCallback:
        cherry_pick_settings = experiment_settings.cherry_pick_settings

        cherry_pick_datasets = DatasetLoader[PairPreferenceDataset](PairPreferenceDataset).load_datasets(
            cherry_pick_settings.dataset_settings, tokenizer=tokenizer, strategy=DatasetStrategy.TRAIN
        )

        metrics = [
            Metric.by_name(metric.type)(MetricSettingsRegistry.by_name(metric.type)(**metric.parameters))
            for metric in cherry_pick_settings.metric_settings
        ]

        return RmCherryPickCallback(
            cherry_pick_settings=cherry_pick_settings,
            datasets=cherry_pick_datasets,
            metrics=metrics,
        )

    @staticmethod
    def _get_training_args(experiment_settings: RMTrainExperimentSettings) -> TrainingArguments:
        return TrainingArguments(
            output_dir=str(experiment_settings.log_path / TRAINER_LOGS_FOLDER),
            label_names=[],
            remove_unused_columns=False,
            **experiment_settings.trainer_settings.dict(),
        )

    @staticmethod
    def _get_trainer(
        training_args: TrainingArguments,
        experiment_settings: RMTrainExperimentSettings,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: DataCollatorMixin,
        **_kwargs,
    ):
        return RMTrainer(
            model=model,
            tokenizer=tokenizer,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            data_collator=data_collator,
            compute_metrics=compute_rm_metrics,
            callbacks=[],
        )

    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: DataCollatorMixin) -> None:
        logger.info(f'Train sample input_ids:\n{dataset[0]}')
        logger.info(f'Train sample example:\n{self.tokenizer.decode(dataset[0]["inputs_w"]["input_ids"])}')
        logger.info(f'Train sample example:\n{self.tokenizer.decode(dataset[0]["inputs_l"]["input_ids"])}')



import os
from pathlib import Path
from typing import Any

import torch
from peft import PeftModel
from torch import nn
from transformers import PreTrainedModel
from transformers.integrations.deepspeed import is_deepspeed_zero3_enabled
from transformers.trainer_pt_utils import nested_detach
from transformers.trainer_utils import PREFIX_CHECKPOINT_DIR
from transformers.utils import logging

from turbo_alignment.trainers.multigpu import MultiGPUCherryPicksTrainer
from turbo_alignment.trainers.utils import concatenated_inputs

logger = logging.get_logger(__name__)


class RMTrainer(MultiGPUCherryPicksTrainer):
    def concatenated_forward(self, model: nn.Module, batch: dict[str, Any]) -> tuple[torch.Tensor, torch.Tensor]:
        concatenated_batch = concatenated_inputs(batch, device=self.accelerator.device)
        all_rewards = model(
            concatenated_batch['input_ids'], attention_mask=concatenated_batch['attention_mask'], return_dict=True
        )[0]

        chosen_idxs = batch['inputs_w']['input_ids'].shape[0]

        chosen_rewards = all_rewards[:chosen_idxs]
        rejected_rewards = all_rewards[chosen_idxs:]

        return chosen_rewards, rejected_rewards

    def compute_loss(self, model, inputs, return_outputs=False) -> tuple[torch.Tensor, dict[str, Any]] | torch.Tensor:
        rewards_w, rewards_l = self.concatenated_forward(model, inputs)

        loss = -torch.nn.functional.logsigmoid(rewards_w - rewards_l).mean()
        if return_outputs:
            return loss, {'rewards_w': rewards_w, 'rewards_l': rewards_l}
        return loss

    def prediction_step(
        self,
        model: PreTrainedModel | nn.Module,
        inputs: dict[str, dict[str, torch.Tensor]],
        prediction_loss_only: bool,
        ignore_keys: list[str] | None,
    ) -> tuple[torch.Tensor, torch.Tensor | None, torch.Tensor | None]:
        inputs = self._prepare_inputs(inputs)
        if ignore_keys is None:
            if hasattr(self.model, 'config'):
                ignore_keys = getattr(self.model.config, 'keys_to_ignore_at_inference', [])
            else:
                ignore_keys = []

        with torch.no_grad():
            loss, logits_dict = self.compute_loss(model, inputs, return_outputs=True)

        if prediction_loss_only:
            return (loss, None, None)

        loss = loss.detach()
        logits = tuple(v for k, v in logits_dict.items() if k not in ignore_keys)
        logits = nested_detach(logits)
        logits = torch.stack(logits).mean(dim=2).T

        labels = logits[:, 0] > logits[:, 1]

        labels = labels.long()

        return loss, logits, labels

    def _save_checkpoint(self, model, trial, metrics=None):
        if isinstance(model, PeftModel) and is_deepspeed_zero3_enabled():
            logger.info('Running custom _save_checkpoint')
            checkpoint_folder = f'{PREFIX_CHECKPOINT_DIR}-{self.state.global_step}'
            run_dir = self._get_output_dir(trial=trial)
            output_dir = Path(os.path.join(run_dir, checkpoint_folder))

            (output_dir / 'cls_head').mkdir(parents=True, exist_ok=True)

            torch.save(model.base_model.model.score.state_dict(), output_dir / 'cls_head' / 'cls_head.pt')

        return super()._save_checkpoint(model=model, trial=trial, metrics=metrics)  # pylint: disable=no-member



from typing import Any

import torch

from turbo_alignment.dataset.chat.models import ChatDatasetRecord
from turbo_alignment.generators.chat import ChatGenerator
from turbo_alignment.settings.generators.outputs.chat import (
    AnswerMessage,
    RagInferenceOutput,
)


class RagGenerator(ChatGenerator):
    def _generate_from_single_record(
        self,
        record: dict[str, Any],
        original_record: ChatDatasetRecord,
        dataset_name: str,
    ) -> RagInferenceOutput:
        input_ids = torch.unsqueeze(record['input_ids'], 0).to(self.device)

        answer_indices, document_indices, doc_scores = self._model.generate(
            inputs=input_ids,
            generation_config=self._transformers_generator_parameters,
            tokenizer=self._tokenizer.current_tokenizer,
            pad_token_id=self._tokenizer.pad_token_id,
        )

        answers = self._decode(token_indices=answer_indices.cpu())
        documents = self._decode(token_indices=document_indices.cpu())
        doc_scores = list(doc_scores[0])

        return RagInferenceOutput(
            id=original_record.id,
            dataset_name=dataset_name,
            messages=original_record.messages,
            label=original_record.label,
            answers=[AnswerMessage(id=str(i), content=a) for i, a in enumerate(answers)],
            documents=documents,
            doc_scores=doc_scores,
        )



from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Tuple

import torch
from safetensors.torch import save_file

from turbo_alignment.common.data.io import write_json
from turbo_alignment.common.data.multimodal.base import BaseModalityReader
from turbo_alignment.common.data.multimodal.registry import ModalityReaderRegistry
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.registry import Params
from turbo_alignment.modeling.multimodal.encoders import ModalityEncoderRegistry
from turbo_alignment.modeling.multimodal.encoders.base import BaseModalityEncoder
from turbo_alignment.pipelines.base import BaseStrategy
from turbo_alignment.settings.datasets.multimodal import (
    MultimodalDatasetProcessingSettings,
)
from turbo_alignment.settings.modality import (
    Modality,
    ModalityEncoderSettings,
    ModalityReaderSettings,
)

logger = get_project_logger()


class PreprocessMultimodalDatasetStrategy(BaseStrategy):
    @staticmethod
    def _load_modality_reader_encoder(
        reader_settings: ModalityReaderSettings, encoder_settings: ModalityEncoderSettings, modality: Modality
    ) -> Tuple[BaseModalityReader, BaseModalityEncoder]:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        reader = ModalityReaderRegistry.by_name(modality).from_params(
            Params({'type': reader_settings.reader_type, 'reader_path': reader_settings.reader_path})
        )
        encoder = ModalityEncoderRegistry.by_name(encoder_settings.modality_encoder_type)(
            encoder_path=encoder_settings.encoder_path
        ).to(device)
        return (reader, encoder)

    def _read_modality_objects(self, reader: BaseModalityReader, dataset_path: Path):
        modality_tensors = []

        available_extensions = ('jpg', 'jpeg', 'png', 'svg')

        total_number_of_objects = len(list(dataset_path.iterdir()))

        logger.info('📖 Reading modality objects...')
        files_paths: list[Path] = []
        for extension in available_extensions:
            files_paths.extend(dataset_path.glob(f'*.{extension}'))

        modality_tensors = self._async_process_files(reader, files_paths, total_number_of_objects)

        modality_tensors = torch.cat(modality_tensors)
        return modality_tensors, files_paths

    @staticmethod
    def _encode_modality_objects(
        modality_tensors: torch.Tensor, encoder: BaseModalityEncoder, batch_size: int
    ) -> torch.Tensor:
        encoded_modality_tensors = []

        logger.info('👩‍💻 Encoding objects...')
        batched_modality_tensors = modality_tensors.split(batch_size)
        for i, batch in enumerate(batched_modality_tensors):
            logger.info(f'👩‍💻 Encoded {i} / {len(batched_modality_tensors)} batches')
            encoded_modality_tensor_batch = encoder.encode(batch.to(encoder.device)).detach().cpu()
            encoded_modality_tensors.append(encoded_modality_tensor_batch)

        encoded_modality_tensors = torch.cat(encoded_modality_tensors)

        return encoded_modality_tensors

    @staticmethod
    def _build_encoder_config(encoder) -> dict:
        return {'emb_dim': encoder.emb_dim}

    @staticmethod
    def _async_process_files(reader, files_paths, total_number_of_objects):
        modality_tensors = [None] * len(files_paths)

        with ThreadPoolExecutor() as executor:
            future_to_index = {
                executor.submit(reader.read, str(file_path)): i for i, file_path in enumerate(files_paths)
            }

            for i, future in enumerate(as_completed(future_to_index)):
                index = future_to_index[future]
                if i % 1000 == 0:
                    logger.info(f'📖 Read {i} / {total_number_of_objects} objects')
                try:
                    modality_tensor = future.result()
                    modality_tensors[index] = modality_tensor
                except Exception as exc:
                    logger.error(f'Error reading file at index {index}: {exc}')

        logger.info(f'📖 Successfully read {len(files_paths)} objects!')
        return modality_tensors

    @staticmethod
    def _get_safetensor_dict(encoded_modality_tensors, encoded_file_paths):
        tensors = {}
        for file, tensor in zip(encoded_file_paths, encoded_modality_tensors):
            tensors[file.name] = tensor.clone().detach()
        return tensors

    def run(self, experiment_settings: MultimodalDatasetProcessingSettings) -> None:
        logger.info(f'👩 Start dataset processing with the following settings:\n{experiment_settings}')
        reader, encoder = self._load_modality_reader_encoder(
            experiment_settings.reader_settings, experiment_settings.encoder_settings, experiment_settings.modality
        )
        modality_tensors, encoded_file_paths = self._read_modality_objects(reader, experiment_settings.dataset_path)
        encoded_modality_tensors = self._encode_modality_objects(
            modality_tensors, encoder, experiment_settings.batch_size
        )

        logger.info('👩 Saving encoded objects...')

        experiment_settings.output_file_path.mkdir(parents=True, exist_ok=True)

        tensors = self._get_safetensor_dict(encoded_modality_tensors, encoded_file_paths)

        del encoded_modality_tensors

        save_file(
            tensors,
            experiment_settings.output_file_path
            / (
                experiment_settings.modality.value
                + '.'
                + experiment_settings.encoder_settings.modality_encoder_type
                + '.safetensors'
            ),
        )

        encoder_config_path = experiment_settings.output_file_path / (
            experiment_settings.modality.value
            + '.'
            + experiment_settings.encoder_settings.modality_encoder_type
            + '.config'
        )

        logger.info(f'👩 Saving encoder config to {encoder_config_path}')
        encoder_config = self._build_encoder_config(encoder)
        write_json(encoder_config, encoder_config_path)



import string

import evaluate
import nltk

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MeteorSettings
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.METEOR)
class MeteorMetric(Metric):
    def __init__(self, settings: MeteorSettings) -> None:
        super().__init__(settings=settings)
        self.compute_element_wise_meteor = settings.compute_element_wise_meteor
        self.element_wise_meteor_label = settings.element_wise_meteor_label
        self.meteor_metric = evaluate.load('meteor')

        nltk.download('wordnet')

    def compute(self, **kwargs) -> list[MetricResults]:
        references: list[str] = kwargs.get('references', None)
        predictions: list[str] = kwargs.get('predictions', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        references_exist = references and len(references) > 0

        if not references_exist:
            return [MetricResults(element_wise_scores=[], need_average=self._settings.need_average)]

        element_wise_scores = []

        if self.compute_element_wise_meteor:
            element_wise_accuracy = self._compute_element_wise(
                references=references,
                predictions=predictions,
                label=dataset_name + '@@' + self.element_wise_meteor_label,
            )

            element_wise_scores.append(element_wise_accuracy)

        return [
            MetricResults(element_wise_scores=element_wise_scores, need_average=need_average)
            for need_average in self._settings.need_average
        ]

    def _compute_element_wise(self, references: list[str], predictions: list[str], label: str) -> ElementWiseScores:
        element_wise_meteor_scores = []
        for reference_batch, prediction_batch in zip(references, predictions):
            for reference, prediction in zip(reference_batch, prediction_batch):
                if len(reference) > 0 and len(prediction) > 0:
                    score = self._compute_meteor(reference=reference, prediction=prediction)
                else:
                    score = 0
                element_wise_meteor_scores.append(score)

        return ElementWiseScores(label=label, values=element_wise_meteor_scores)

    def _compute_meteor(self, reference: str, prediction: str) -> float:
        return self.meteor_metric.compute(
            predictions=[self._filter_punctuatuion(prediction)], references=[self._filter_punctuatuion(reference)]
        )['meteor']

    @staticmethod
    def _filter_punctuatuion(s: str) -> str:
        if s:
            return ''.join(char for char in s if char not in string.punctuation)
        return s



from abc import abstractmethod
from typing import Generic, TypeVar

from turbo_alignment.common.data.io import write_jsonl
from turbo_alignment.dataset.sampling.models import SamplingDatasetRecord
from turbo_alignment.pipelines.base import BaseStrategy
from turbo_alignment.settings.pipelines.sampling import BaseSamplingSettings

SamplingSettingsT = TypeVar('SamplingSettingsT', bound=BaseSamplingSettings)


class BaseSamplingStrategy(BaseStrategy, Generic[SamplingSettingsT]):
    @abstractmethod
    def sample(self, experiment_settings: SamplingSettingsT) -> list[SamplingDatasetRecord]:
        ...

    def run(self, experiment_settings: SamplingSettingsT) -> None:
        experiment_settings.save_path.mkdir(parents=True, exist_ok=True)

        sampled_outputs: list[SamplingDatasetRecord] = self.sample(experiment_settings)

        write_jsonl([out.dict() for out in sampled_outputs], experiment_settings.save_path / 'sampled.jsonl')



from abc import ABC, abstractmethod

from turbo_alignment.common.logging.clearml import create_clearml_task
from turbo_alignment.common.logging.weights_and_biases import create_wandb_run
from turbo_alignment.common.registry import Registrable
from turbo_alignment.common.tf.callbacks.logging import (
    ClearMLLoggingCallback,
    LoggingCallback,
    WandbLoggingCallback,
)
from turbo_alignment.settings.logging.clearml import ClearMLSettings
from turbo_alignment.settings.logging.weights_and_biases import WandbSettings
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings


class LoggingRegistry(Registrable):
    ...


class LoggingMixin(ABC):
    @staticmethod
    @abstractmethod
    def get_logging_callback(experiment_settings: BaseTrainExperimentSettings) -> LoggingCallback:
        ...


@LoggingRegistry.register(WandbSettings.__name__)
class WandbLoggingMixin(LoggingMixin):
    @staticmethod
    def get_logging_callback(experiment_settings: BaseTrainExperimentSettings) -> WandbLoggingCallback:
        logging_settings: WandbSettings = WandbSettings(**experiment_settings.logging_settings.dict())
        wandb_run = create_wandb_run(parameters=logging_settings, config=experiment_settings.dict())
        return WandbLoggingCallback(wandb_run=wandb_run)


@LoggingRegistry.register(ClearMLSettings.__name__)
class ClearMLLoggingMixin(LoggingMixin):
    @staticmethod
    def get_logging_callback(experiment_settings: BaseTrainExperimentSettings) -> ClearMLLoggingCallback:
        logging_settings: ClearMLSettings = ClearMLSettings(**experiment_settings.logging_settings.dict())
        clearml_task = create_clearml_task(parameters=logging_settings, config=experiment_settings.dict())
        return ClearMLLoggingCallback(clearml_task=clearml_task)



from pydantic import BaseModel, field_validator


class DatasetRecord(BaseModel):
    id: str

    @field_validator('id', mode='before')
    def convert_int_id_to_str(cls, values: int) -> str:
        if isinstance(values, int):
            return str(values)
        return values



from typing import Callable, cast

from torch.utils.data import Dataset
from transformers import PreTrainedModel, PreTrainedTokenizerBase, TrainingArguments
from transformers.data.data_collator import DataCollatorMixin, DataCollatorWithPadding

from turbo_alignment.cherry_picks.classification import ClassificationCherryPickCallback
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.constants import TRAINER_LOGS_FOLDER
from turbo_alignment.dataset.classification.classification import ClassificationDataset
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MetricSettingsRegistry
from turbo_alignment.pipelines.train.base import BaseTrainStrategy
from turbo_alignment.settings.datasets.base import DatasetStrategy
from turbo_alignment.settings.datasets.classification import (
    ClassificationMultiDatasetSettings,
)
from turbo_alignment.settings.pipelines.train.classification import (
    ClassificationTrainExperimentSettings,
)
from turbo_alignment.trainers.classification import (
    ClassificationTrainer,
    auto_class_weights,
)

logger = get_project_logger()


class TrainClassificationStrategy(BaseTrainStrategy[ClassificationTrainExperimentSettings]):
    @staticmethod
    def _get_data_collator(
        experiment_settings: ClassificationTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **_kwargs,
    ) -> Callable:
        dataset_settings: ClassificationMultiDatasetSettings = cast(
            ClassificationMultiDatasetSettings, experiment_settings.train_dataset_settings
        )
        return DataCollatorWithPadding(tokenizer=tokenizer, max_length=dataset_settings.chat_settings.max_tokens_count)

    @staticmethod
    def _get_cherry_pick_callback(
        experiment_settings: ClassificationTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **_kwargs,
    ) -> ClassificationCherryPickCallback:
        cherry_pick_settings = experiment_settings.cherry_pick_settings

        cherry_pick_datasets = DatasetLoader[ClassificationDataset](ClassificationDataset).load_datasets(
            cherry_pick_settings.dataset_settings, tokenizer=tokenizer, strategy=DatasetStrategy.INFERENCE
        )

        metrics = [
            Metric.by_name(metric.type)(MetricSettingsRegistry.by_name(metric.type)(**metric.parameters))
            for metric in cherry_pick_settings.metric_settings
        ]

        return ClassificationCherryPickCallback(
            cherry_pick_settings=cherry_pick_settings,
            datasets=cherry_pick_datasets,
            metrics=metrics,
        )

    @staticmethod
    def _get_training_args(experiment_settings: ClassificationTrainExperimentSettings) -> TrainingArguments:
        return TrainingArguments(
            output_dir=str(experiment_settings.log_path / TRAINER_LOGS_FOLDER),
            label_names=['labels'],
            remove_unused_columns=False,
            **experiment_settings.trainer_settings.dict(exclude={'loss_settings'}),
        )

    @staticmethod
    def _get_trainer(
        training_args: TrainingArguments,
        experiment_settings: ClassificationTrainExperimentSettings,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: DataCollatorMixin,
    ):
        if experiment_settings.trainer_settings.loss_settings.alpha == 'auto':
            experiment_settings.trainer_settings.loss_settings.alpha = auto_class_weights(train_dataset)
            logger.info(f'Auto computed class weights: {experiment_settings.trainer_settings.loss_settings.alpha}')

        return ClassificationTrainer(
            model=model,
            tokenizer=tokenizer,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            data_collator=data_collator,
            callbacks=[],
            loss_settings=experiment_settings.trainer_settings.loss_settings,
        )

    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: DataCollatorMixin) -> None:
        logger.info('Input check: {input_ids}'.format(input_ids=collator([dataset[0], dataset[1]])['input_ids'][0]))

        logger.info('Label check: {labels}'.format(labels=collator([dataset[0], dataset[1]])['labels'][0]))



from typing import Any

import torch
from transformers import DataCollatorWithPadding, PreTrainedTokenizerBase

from turbo_alignment.dataset.classification.models import ClassificationDatasetRecord
from turbo_alignment.generators.base import BaseGenerator
from turbo_alignment.settings.generators.outputs.classification import (
    ClassificationInferenceOutput,
)


class ClassificationGenerator(BaseGenerator[ClassificationDatasetRecord, ClassificationInferenceOutput]):
    def __init__(self, tokenizer: PreTrainedTokenizerBase, **kwargs):
        super().__init__(tokenizer=tokenizer, **kwargs)

        self._collator = DataCollatorWithPadding(tokenizer=tokenizer, padding=True)

    def _generate_from_batch(
        self, records: list[dict[str, Any]], original_records: list[ClassificationDatasetRecord], dataset_name: str
    ) -> list[ClassificationInferenceOutput]:
        inputs = [{'input_ids': record['input_ids']} for record in records]

        with torch.no_grad():
            input_ids = self._collator(inputs)['input_ids']
            input_ids = input_ids.to(self.device)
            output_logits = self._model(input_ids).logits
            classes = torch.argmax(output_logits, dim=1)

        return [
            ClassificationInferenceOutput(
                id=record.id,
                messages=record.messages,
                predicted_label=cl.item(),
                dataset_name=dataset_name,
            )
            for record, cl in zip(original_records, classes)
        ]



from pathlib import Path
from typing import Any

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class TrainerSettings(ExtraFieldsNotAllowedBaseModel):
    evaluation_strategy: str = 'steps'
    save_strategy: str = 'steps'
    per_device_train_batch_size: int = 4
    per_device_eval_batch_size: int = 4
    gradient_accumulation_steps: int = 32
    eval_steps: int = 150
    save_steps: int = 150
    logging_steps: int = 5
    learning_rate: float = 0.0003
    num_train_epochs: int = 3
    max_steps: int = -1
    lr_scheduler_type: str = 'cosine'
    lr_scheduler_kwargs: dict[str, Any] = {}
    warmup_steps: int = 0
    warmup_ratio: float = 0.0
    fp16: bool = True
    bf16: bool = False
    tf32: bool = False
    torch_compile: bool = False
    optim: str = 'adamw_torch'
    adam_beta1: float = 0.9
    adam_beta2: float = 0.999
    adam_epsilon: float = 1e-8
    weight_decay: float = 0.0
    max_grad_norm: float = 1.0
    deepspeed: Path | None = None
    save_total_limit: int = 1
    save_only_model: bool = False
    no_cuda: bool = False
    prediction_loss_only: bool = False
    load_best_model_at_end: bool = True
    logging_first_step: bool = True
    fsdp_config: dict[str, Any] | None = None
    fsdp: str | list[str] | None = ''
    dataloader_num_workers: int = 8
    dataloader_prefetch_factor: int | None = None
    dataloader_persistent_workers: bool | None = False
    dataloader_pin_memory: bool | None = True
    gradient_checkpointing: bool = False
    gradient_checkpointing_kwargs: dict[str, Any] = {}
    neftune_noise_alpha: float | None = None
    report_to: list[str] = []



from turbo_alignment.settings.cherry_pick import ChatCherryPickSettings
from turbo_alignment.settings.datasets.chat import ChatMultiDatasetSettings
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings


class SftTrainExperimentSettings(BaseTrainExperimentSettings):
    train_dataset_settings: ChatMultiDatasetSettings
    val_dataset_settings: ChatMultiDatasetSettings

    cherry_pick_settings: ChatCherryPickSettings



from pathlib import Path
from typing import Literal

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.datasets.base import (
    BaseDatasetSettings,
    DatasetType,
    MultiDatasetSettings,
)
from turbo_alignment.settings.datasets.chat import ChatPromptTemplate
from turbo_alignment.settings.modality import (
    Modality,
    ModalityEncoderSettings,
    ModalityReaderSettings,
)


class MultimodalDatasetSettings(BaseDatasetSettings):
    dataset_type: Literal[DatasetType.MULTIMODAL] = DatasetType.MULTIMODAL
    max_tokens_count: int
    only_answer_loss: bool
    prompt_template: ChatPromptTemplate
    modality_token_mapping: dict[Modality, str]
    modality_reader_settings_mapping: dict[Modality, ModalityReaderSettings | None]
    start_modality_token: str
    end_modality_token: str
    n_modality_embeddings: int
    truncate_top: bool
    random_cut: bool = False
    keep_end: bool | None = None
    only_last_replica_loss: bool = False
    ignore_system_prompt: bool = False


class MultimodalMultiDatasetSettings(MultimodalDatasetSettings, MultiDatasetSettings):
    ...


class MultimodalDatasetProcessingSettings(ExtraFieldsNotAllowedBaseModel):
    reader_settings: ModalityReaderSettings
    encoder_settings: ModalityEncoderSettings
    modality: Modality
    dataset_path: Path
    batch_size: int = 64
    output_file_path: Path



import torch
from transformers import PreTrainedTokenizerBase

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.utils import calculate_cross_entropy
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.PERPLEXITY)
class PerplexityMetric(Metric):
    def compute(self, **kwargs) -> list[MetricResults]:
        tokenizer: PreTrainedTokenizerBase = kwargs.get('tokenizer', None)
        logits: torch.Tensor = kwargs.get('logits', None)
        labels: torch.Tensor = kwargs.get('answer_tokens_ids', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if logits is None:
            raise ValueError('logits should not be None')

        if labels is None:
            raise ValueError('labels should not be None')

        if tokenizer is None:
            raise ValueError('tokenizer should not be None')

        perplexity_values = []
        for logit, answer in zip(logits, labels):
            perplexity_values.extend(self.calculate_perplexity(logit, answer, tokenizer.pad_token_id))

        return [
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(
                        label=dataset_name + '@@' + 'perplexity',
                        values=perplexity_values,
                    )
                ],
                need_average=need_average,
            )
            for need_average in self._settings.need_average
        ]

    @staticmethod
    def calculate_perplexity(logits: torch.Tensor, labels: torch.Tensor, pad_token_id: int) -> list[float]:
        ppl = calculate_cross_entropy(logits, labels, pad_token_id, reduction='none')
        return torch.exp(ppl).detach().cpu().tolist()



import time

import numpy as np
import torch
from transformers.tokenization_utils import PreTrainedTokenizer
from transformers.tokenization_utils_base import BatchEncoding

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.modeling.rag.rag_index import CustomHFIndex
from turbo_alignment.modeling.rag.rag_tokenizer import RagTokenizer
from turbo_alignment.settings.rag_model import RAGPreTrainedModelSettings

logger = get_project_logger()


class RagRetriever:
    """
    Stores indexed documents, retrieves documents relevant for a query.
    Note: __call__ returns documents encoded with generator tokenizer.
    """

    def __init__(self, config: RAGPreTrainedModelSettings, rag_tokenizer: RagTokenizer, init_retrieval=True) -> None:
        self._init_retrieval = init_retrieval
        super().__init__()

        self.index = self._build_index(config)

        self.generator_tokenizer = rag_tokenizer.generator
        self.question_encoder_tokenizer = rag_tokenizer.question_encoder
        self.rag_tokenizer = rag_tokenizer

        self.n_docs = config.retrieval_settings.n_docs
        self.batch_size = config.retrieval_settings.retrieval_batch_size

        self.config: RAGPreTrainedModelSettings = config
        if self._init_retrieval:
            self.init_retrieval()

        self.ctx_encoder_tokenizer = None
        self.return_tokenized_docs = False

    @staticmethod
    def _build_index(config) -> CustomHFIndex:
        # CanonicalHFIndex is removed, return if needed
        return CustomHFIndex.load_from_disk(
            vector_size=config.retrieval_settings.retrieval_vector_size,
            dataset_path=config.index_settings.passages_path,
            index_path=config.index_settings.index_path,
        )

    def save_pretrained(self, save_directory) -> None:
        # self.config.save_pretrained(save_directory) FIXME
        self.rag_tokenizer.save_pretrained(save_directory)

    def init_retrieval(self) -> None:
        logger.info('initializing retrieval')
        self.index.init_index()

    def postprocess_docs(self, docs: list[dict], n_docs: int, return_tensors: str = 'pt'):
        def cat_input_and_doc(doc_title, doc_text):
            if doc_title.startswith('"'):
                doc_title = doc_title[1:]
            if doc_title.endswith('"'):
                doc_title = doc_title[:-1]
            out = (
                doc_title
                + self.config.retrieval_settings.title_sep
                + doc_text
                + self.config.retrieval_settings.doc_sep
            ).replace('  ', ' ')
            return out.strip()

        rag_input_strings = [
            cat_input_and_doc(docs[i]['title'][j], docs[i]['text'][j]) for i in range(len(docs)) for j in range(n_docs)
        ]

        contextualized_inputs = self.generator_tokenizer.batch_encode_plus(
            rag_input_strings,
            max_length=self.config.retrieval_settings.max_doc_length,
            return_tensors=return_tensors,
            padding=True,
            truncation=True,
        )

        return contextualized_inputs['input_ids'], contextualized_inputs['attention_mask']

    def _chunk_tensor(self, t, chunk_size):
        return [t[i : i + chunk_size] for i in range(0, len(t), chunk_size)]

    def _main_retrieve(self, question_hidden_states: np.ndarray, n_docs: int) -> tuple[np.ndarray, np.ndarray]:
        question_hidden_states_batched = self._chunk_tensor(question_hidden_states, self.batch_size)
        ids_batched: list[np.ndarray] = []
        vectors_batched: list[np.ndarray] = []
        for question_hidden_states_batch in question_hidden_states_batched:
            start_time = time.time()
            ids, vectors = self.index.get_top_docs(question_hidden_states_batch, n_docs)
            logger.debug(
                f'index search time: {time.time() - start_time} sec, batch size {question_hidden_states_batch.shape}'
            )
            ids_batched.extend(ids)
            vectors_batched.extend(vectors)
        return (
            np.array(ids_batched),
            np.array(vectors_batched),
        )  # shapes (batch_size, n_docs) and (batch_size, n_docs, d)

    def retrieve(self, question_hidden_states: np.ndarray, n_docs: int) -> tuple[np.ndarray, np.ndarray, list[dict]]:
        doc_ids, retrieved_doc_embeds = self._main_retrieve(question_hidden_states, n_docs)
        return retrieved_doc_embeds, doc_ids, self.index.get_doc_dicts(doc_ids)

    def set_ctx_encoder_tokenizer(self, ctx_encoder_tokenizer: PreTrainedTokenizer) -> None:
        # used in end2end retriever training
        self.ctx_encoder_tokenizer = ctx_encoder_tokenizer
        self.return_tokenized_docs = True

    def __call__(
        self,
        question_input_ids: torch.Tensor,
        question_hidden_states: np.ndarray,
        n_docs: int,
        return_tensors=None,
    ) -> BatchEncoding:
        retrieved_doc_embeds, doc_ids, docs = self.retrieve(question_hidden_states, n_docs)

        context_input_ids, context_attention_mask = self.postprocess_docs(docs, n_docs, return_tensors=return_tensors)
        return BatchEncoding(
            {
                'context_input_ids': context_input_ids,
                'context_attention_mask': context_attention_mask,
                'retrieved_doc_embeds': retrieved_doc_embeds,
                'doc_ids': doc_ids,
            },
            tensor_type=return_tensors,
        )



from pathlib import Path

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.datasets.sampling import SamplingDatasetSettings
from turbo_alignment.settings.model import (
    PreTrainedAdaptersModelSettings,
    PreTrainedModelSettings,
)
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings


class BaseSamplingSettings(ExtraFieldsNotAllowedBaseModel):
    dataset_settings: SamplingDatasetSettings

    save_path: Path
    N: int


class RandomSamplingSettings(BaseSamplingSettings):
    ...


class BaseSamplingWithRMSettings(BaseSamplingSettings):
    rm: PreTrainedAdaptersModelSettings | PreTrainedModelSettings
    tokenizer_settings: TokenizerSettings

    rm_batch_size: int


class SamplingWithRMSettings(BaseSamplingWithRMSettings):
    ...


class RSOSamplingSettings(BaseSamplingWithRMSettings):
    beta: float



# mypy: ignore-errors
# pylint: skip-file
import torch
import triton
import triton.language as tl
from torch.nn import CrossEntropyLoss


@triton.jit
def liger_cross_entropy_kernel(
    X_ptr,
    X_stride,
    Y_ptr,
    Y_stride,
    loss_ptr,
    loss_stride,
    n_cols,
    n_non_ignore,
    ignore_index,
    label_smoothing: tl.constexpr,
    reduction: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    program_id = tl.program_id(0).to(tl.int64)

    Y_ptr += program_id * Y_stride
    y = tl.load(Y_ptr)

    X_ptr += program_id * X_stride

    if y == ignore_index:
        for i in range(0, n_cols, BLOCK_SIZE):
            X_offsets = i + tl.arange(0, BLOCK_SIZE)
            tl.store(X_ptr + X_offsets, 0.0, mask=X_offsets < n_cols)
        return

    loss_ptr += program_id * loss_stride

    m = float('-inf')
    d = 0.0
    ori_X_y = tl.load(X_ptr + y)

    scaled_x_sum = 0.0
    eps = label_smoothing / n_cols

    for i in range(0, n_cols, BLOCK_SIZE):
        X_offsets = i + tl.arange(0, BLOCK_SIZE)
        X_block = tl.load(X_ptr + X_offsets, mask=X_offsets < n_cols, other=float('-inf'))
        block_max = tl.max(X_block)
        if label_smoothing > 0:
            scaled_x_sum += tl.sum(tl.where(X_offsets < n_cols, -eps * X_block, 0.0))
        m_new = tl.maximum(m, block_max)
        d = d * tl.exp(m - m_new) + tl.sum(tl.exp(X_block - m_new))
        m = m_new

    for i in range(0, n_cols, BLOCK_SIZE):
        X_offsets = i + tl.arange(0, BLOCK_SIZE)
        X_block = tl.load(X_ptr + X_offsets, mask=X_offsets < n_cols, other=float('-inf'))
        if reduction == 'mean':
            X_block = (tl.exp(X_block - m) / d - eps) / (n_non_ignore)
        else:
            X_block = tl.exp(X_block - m) / d - eps

        tl.store(X_ptr + X_offsets, X_block, mask=X_offsets < n_cols)

    tl.debug_barrier()

    loss = -(ori_X_y - m - tl.log(d))

    if label_smoothing > 0:
        smooth_loss = scaled_x_sum + label_smoothing * (m + tl.log(d))
        loss = loss * (1 - label_smoothing) + smooth_loss

    if reduction == 'mean':
        loss = loss / n_non_ignore

    X_y = tl.load(X_ptr + y)
    if reduction == 'mean':
        X_y += -(1 - label_smoothing) / (n_non_ignore)
    else:
        X_y += -(1 - label_smoothing)

    tl.store(loss_ptr, loss)
    tl.store(X_ptr + y, X_y)


MAX_FUSED_SIZE = 65536 // 2


@triton.jit
def element_mul_kernel(
    X_ptr,
    X_stride,
    grad_output_ptr,
    n_cols,
    BLOCK_SIZE: tl.constexpr,
):
    program_id = tl.program_id(0).to(tl.int64)

    X_ptr += program_id * X_stride

    grad_output = tl.load(grad_output_ptr)

    for i in range(0, n_cols, BLOCK_SIZE):
        X_offsets = i + tl.arange(0, BLOCK_SIZE)
        X_block = tl.load(X_ptr + X_offsets, mask=X_offsets < n_cols)
        tl.store(X_ptr + X_offsets, X_block * grad_output, mask=X_offsets < n_cols)


def cross_entropy_forward(_input, target, ignore_index, label_smoothing, reduction):
    BT, V = _input.shape
    n_rows = BT

    BLOCK_SIZE = min(MAX_FUSED_SIZE, triton.next_power_of_2(V))

    loss_1d = torch.zeros(n_rows, dtype=_input.dtype, device=_input.device)

    n_non_ignore = (target != ignore_index).sum().item()

    if _input.stride(-1) != 1:
        _input = _input.contiguous()
    if target.stride(-1) != 1:
        target = target.contiguous()

    liger_cross_entropy_kernel[(n_rows,)](
        X_ptr=_input,
        X_stride=_input.stride(-2),
        Y_ptr=target,
        Y_stride=target.stride(-1),
        loss_ptr=loss_1d,
        loss_stride=loss_1d.stride(-1),
        n_cols=V,
        n_non_ignore=n_non_ignore,
        ignore_index=ignore_index,
        label_smoothing=label_smoothing,
        reduction=reduction,
        BLOCK_SIZE=BLOCK_SIZE,
        num_warps=32,
    )

    loss = torch.sum(loss_1d)
    return loss, _input


def cross_entropy_backward(_input, grad_output):
    if torch.equal(grad_output, torch.tensor(1.0, device=grad_output.device)):
        pass

    else:
        BT, V = _input.shape
        n_rows = BT
        BLOCK_SIZE = min(MAX_FUSED_SIZE, triton.next_power_of_2(V))

        element_mul_kernel[(n_rows,)](
            _input,
            _input.stride(-2),
            grad_output,
            V,
            BLOCK_SIZE=BLOCK_SIZE,
            num_warps=32,
        )

    return _input


class LigerCrossEntropyFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, _input, target, ignore_index=-100, label_smoothing=0.0, reduction='mean'):
        loss, _input = cross_entropy_forward(_input, target, ignore_index, label_smoothing, reduction)
        ctx.save_for_backward(_input.detach())
        return loss

    @staticmethod
    def backward(ctx, grad_output):
        (_input,) = ctx.saved_tensors
        _input = cross_entropy_backward(_input, grad_output)
        return (
            _input,
            None,
            None,
            None,
            None,
        )


class LigerCrossEntropyLoss(CrossEntropyLoss):
    def __init__(self, *args, **kwargs):
        super(LigerCrossEntropyLoss, self).__init__(*args, **kwargs)
        assert (self.label_smoothing >= 0) and (
            self.label_smoothing <= 1
        ), f'label_smoothing must be between 0.0 and 1.0. Got: {self.label_smoothing}'
        assert self.reduction in {
            'mean',
            'sum',
            'none',
        }, f"reduction must be one of 'mean', 'sum', or 'none'. Got: {self.reduction}"

    def forward(self, _input, target):
        return LigerCrossEntropyFunction.apply(_input, target, self.ignore_index, self.label_smoothing, self.reduction)



from typing import Literal

from turbo_alignment.settings.datasets.base import DatasetType, MultiDatasetSettings
from turbo_alignment.settings.datasets.chat import ChatDatasetSettings


class SamplingDatasetSettings(MultiDatasetSettings):
    dataset_type: Literal[DatasetType.SAMPLING] = DatasetType.SAMPLING
    chat_dataset: ChatDatasetSettings



from turbo_alignment.dataset.chat.models import ChatMessage
from turbo_alignment.settings.generators.outputs.base import BaseInferenceOutput


class ClassificationInferenceOutput(BaseInferenceOutput):
    id: str
    messages: list[ChatMessage]
    predicted_label: int



from enum import Enum
from pathlib import Path

from pydantic import BaseModel

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class ImageBindModelSettings(ExtraFieldsNotAllowedBaseModel):
    pass


class ModalityType(str, Enum):
    VISION = 'vision'
    TEXT = 'text'
    AUDIO = 'audio'


class ImageBindArchitectureSettings(BaseModel):
    vision_embed_dim: int = 1280
    vision_num_blocks: int = 32
    vision_num_heads: int = 16

    rgbt_kernel_size: tuple[int, int, int] = (2, 14, 14)
    rgbt_stride: tuple[int, int, int] = (2, 14, 14)

    text_embed_dim: int = 1024
    text_num_blocks: int = 24
    text_num_heads: int = 16

    audio_kernel_size: int = 16
    audio_num_blocks: int = 12
    audio_num_heads: int = 12
    audio_embed_dim: int = 768
    audio_drop_path_rate: float = 0.1

    audio_stride: int = 10
    audio_num_mel_bins: int = 128
    audio_target_len: int = 204

    out_embed_dim: int = 1024


class ImageBindSettings(BaseModel):
    architecture_settings: ImageBindArchitectureSettings = ImageBindArchitectureSettings()

    weights_path: Path | None = Path('.checkpoints/imagebind_huge.pth')
    is_trainable: bool = False



import numpy as np
from datasets import Dataset, load_from_disk

from turbo_alignment.common.logging import get_project_logger

logger = get_project_logger()


class Index:
    """
    A base class for the Indices encapsulated by the [`RagRetriever`].
    """

    def get_doc_dicts(self, doc_ids: np.ndarray) -> list[dict]:
        """
        Returns a list of dictionaries, containing titles and text of the retrieved documents.

        Args:
            doc_ids (`np.ndarray` of shape `(batch_size, n_docs)`):
                A tensor of document indices.
        """
        raise NotImplementedError

    def get_top_docs(self, question_hidden_states: np.ndarray, n_docs=5) -> tuple[np.ndarray, np.ndarray]:
        """
        For each query in the batch, retrieves `n_docs` documents.

        Args:
            question_hidden_states (`np.ndarray` of shape `(batch_size, vector_size)`):
                An array of query vectors.
            n_docs (`int`):
                The number of docs retrieved per query.

        Returns:
            `np.ndarray` of shape `(batch_size, n_docs)`: A tensor of indices of retrieved documents. `np.ndarray` of
            shape `(batch_size, vector_size)`: A tensor of vector representations of retrieved documents.
        """
        raise NotImplementedError

    def is_initialized(self):
        """
        Returns `True` if index is already initialized.
        """
        raise NotImplementedError

    def init_index(self):
        """
        A function responsible for loading the index into memory. Should be called only once per training run of a RAG
        model. E.g. if the model is trained on multiple GPUs in a distributed setup, only one of the workers will load
        the index.
        """
        raise NotImplementedError


class HFIndexBase(Index):
    def __init__(self, vector_size, dataset, index_initialized=False) -> None:
        self.vector_size = vector_size
        self.dataset = dataset
        self._index_initialized = index_initialized
        self._check_dataset_format(with_index=index_initialized)
        dataset.set_format('numpy', columns=['embeddings'], output_all_columns=True, dtype='float32')

    def _check_dataset_format(self, with_index: bool) -> None:
        if not isinstance(self.dataset, Dataset):
            raise ValueError(f'Dataset should be a datasets.Dataset object, but got {type(self.dataset)}')
        if len({'title', 'text', 'embeddings'} - set(self.dataset.column_names)) > 0:
            raise ValueError(
                'Dataset should be a dataset with the following columns: '
                'title (str), text (str) and embeddings (arrays of dimension vector_size), '
                f'but got columns {self.dataset.column_names}'
            )
        if with_index and 'embeddings' not in self.dataset.list_indexes():
            raise ValueError(
                'Missing faiss index in the dataset. Make sure you called `dataset.add_faiss_index` to compute it '
                'or `dataset.load_faiss_index` to load one from the disk.'
            )

    def init_index(self):
        raise NotImplementedError()

    def is_initialized(self):
        return self._index_initialized

    def get_doc_dicts(self, doc_ids: np.ndarray) -> list[dict]:
        return [self.dataset[doc_ids[i].tolist()] for i in range(doc_ids.shape[0])]

    def get_top_docs(self, question_hidden_states: np.ndarray, n_docs=5) -> tuple[np.ndarray, np.ndarray]:
        _, ids = self.dataset.search_batch('embeddings', question_hidden_states, n_docs)
        docs = [self.dataset[[i for i in indices if i >= 0]] for indices in ids]
        vectors = [doc['embeddings'] for doc in docs]
        for i, vec in enumerate(vectors):
            if len(vec) < n_docs:
                vectors[i] = np.vstack([vec, np.zeros((n_docs - len(vec), self.vector_size))])
        return np.array(ids), np.array(vectors)  # shapes (batch_size, n_docs) and (batch_size, n_docs, d)


class CustomHFIndex(HFIndexBase):
    def __init__(self, vector_size: int, dataset, index_path=None) -> None:
        super().__init__(vector_size, dataset, index_initialized=index_path is None)
        self.index_path = index_path

    @classmethod
    def load_from_disk(cls, vector_size, dataset_path, index_path):
        logger.info(f'Loading passages from {dataset_path}')
        if dataset_path is None or index_path is None:
            raise ValueError(
                'Please provide `dataset_path` and `index_path` after calling `dataset.save_to_disk(dataset_path)` '
                "and `dataset.get_index('embeddings').save(index_path)`."
            )
        dataset = load_from_disk(str(dataset_path))
        return cls(vector_size=vector_size, dataset=dataset, index_path=str(index_path))

    def init_index(self) -> None:
        if not self.is_initialized():
            logger.info(f'Loading index from {self.index_path}')
            self.dataset.load_faiss_index('embeddings', file=self.index_path)
            self._index_initialized = True



from dataclasses import dataclass
from typing import Any, Iterable

import torch
import transformers
from transformers import PreTrainedTokenizerBase

from turbo_alignment.constants import DISABLE_LOSS_LABEL


@dataclass
class PairPreferenceDataCollator:
    tokenizer: PreTrainedTokenizerBase
    add_labels: bool = True
    pad_to_multiple_of: int | None = None

    def _get_batch(
        self, examples: list[dict[str, dict[str, Any]]], tokenizer: PreTrainedTokenizerBase, key: str, max_length: int
    ) -> transformers.BatchEncoding:
        features = [ex[key] for ex in examples]
        labels = [v.tolist() for feature in features for k, v in feature.items() if k == 'labels']
        no_labels_features = [
            {k: v for k, v in feature.items() if k not in ['labels', 'precomputed_margin']} for feature in features
        ]

        batch = tokenizer.pad(
            no_labels_features,
            padding='max_length',
            max_length=max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors='pt',
        )
        if self.add_labels:
            batch['labels'] = torch.tensor(
                [label + (max_length - len(label)) * [DISABLE_LOSS_LABEL] for label in labels]
            )
        return batch

    def __call__(self, examples: list[dict[str, dict[str, Any]]]) -> dict[str, Any]:
        max_length = 0
        for ex in examples:
            for t in ex:
                if isinstance(ex[t], Iterable):
                    if 'input_ids' in ex[t]:
                        max_length = max(max_length, len(ex[t]['input_ids']))

        batch: dict[str, Any] = {
            'inputs_w': dict(self._get_batch(examples, self.tokenizer, 'inputs_w', max_length)),
            'inputs_l': dict(self._get_batch(examples, self.tokenizer, 'inputs_l', max_length)),
        }

        if 'precomputed_margin' in examples[0] and examples[0]['precomputed_margin'] is not None:
            batch['precomputed_margin'] = torch.tensor([ex['precomputed_margin'] for ex in examples])

        return batch



from collections import defaultdict

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.DIST_N)
class DistinctnessMetric(Metric):
    def compute(self, **kwargs) -> list[MetricResults]:
        predictions: list[list[str]] = kwargs.get('predictions', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if predictions is None:
            raise ValueError('predictions should not be None')

        dist_n = defaultdict(list)
        for prompt_answers in predictions:
            ans_dist_n = self.distinctness(prompt_answers)
            for label, value in ans_dist_n.items():
                dist_n[label].append(value)

        return [
            MetricResults(
                element_wise_scores=[
                    ElementWiseScores(label=dataset_name + '@@' + label, values=value)
                    for label, value in dist_n.items()
                ],
                need_average=need_average,
            )
            for need_average in self._settings.need_average
        ]

    @staticmethod
    def distinctness(answers: list[str]) -> dict[str, float]:
        unigrams, bigrams, trigrams = set(), set(), set()
        total_words = 0

        for answer in answers:
            words = answer.split(' ')
            total_words += len(words)
            unigrams.update(words)
            for i in range(len(words) - 1):
                bigrams.add(words[i] + '_' + words[i + 1])
            for i in range(len(words) - 2):
                trigrams.add(words[i] + '_' + words[i + 1] + '_' + words[i + 2])

        return {
            'dist_1': len(unigrams) / total_words,
            'dist_2': len(bigrams) / total_words,
            'dist_3': len(trigrams) / total_words,
        }



from abc import abstractmethod
from collections import defaultdict
from copy import deepcopy
from typing import Any

import torch
from accelerate import Accelerator
from accelerate.utils import is_deepspeed_available
from torch import nn
from transformers import PreTrainedModel

from turbo_alignment.common.registry import Registrable


def concatenated_inputs(
    batch: dict[str, Any], prefix: str = '', device: torch.device = torch.device('cpu')
) -> dict[str, torch.Tensor]:
    """
    Объединяем chosen и rejected инпуты в один батч, чтобы прогнать их за один forward
    """

    grouped_batch: dict[str, list[torch.Tensor]] = defaultdict(list)
    no_grouped_batch_items: dict[str, Any] = {}
    for outcome_key, outcome_inputs in batch.items():
        if outcome_key.startswith(prefix):
            if isinstance(outcome_inputs, dict):
                for k, v in outcome_inputs.items():
                    grouped_batch[k].append(v)
            else:
                no_grouped_batch_items[outcome_key] = outcome_inputs

    concatenated_batch: dict[str, torch.Tensor] = {}
    for k, v in grouped_batch.items():
        concatenated_batch[k] = torch.cat(v, dim=0).to(device)

    for k, v in no_grouped_batch_items.items():
        if isinstance(v, torch.Tensor):
            v = v.to(device)
        concatenated_batch[k] = v

    return concatenated_batch


def prepare_model(
    model: PreTrainedModel | nn.Module, accelerator: Accelerator, is_deepspeed_enabled: bool = False
) -> PreTrainedModel | nn.Module:
    if is_deepspeed_enabled:
        model = prepare_model_for_deepspeed(model, accelerator)
    else:
        model = accelerator.prepare_model(model, evaluation_mode=True)

    return model


def prepare_model_for_deepspeed(
    model: PreTrainedModel | nn.Module, accelerator: Accelerator
) -> PreTrainedModel | nn.Module:
    if not is_deepspeed_available():
        raise ValueError('Deepspeed is not installed')

    import deepspeed

    deepspeed_plugin = accelerator.state.deepspeed_plugin
    config_kwargs = deepcopy(deepspeed_plugin.deepspeed_config)
    if model is not None:
        if hasattr(model, 'config'):
            hidden_size: int | None = (
                max(model.config.hidden_sizes)
                if getattr(model.config, 'hidden_sizes', None)
                else getattr(model.config, 'hidden_size', None)
            )

            if hidden_size is not None and config_kwargs['zero_optimization']['stage'] == 3:
                config_kwargs.update(
                    {
                        'zero_optimization.reduce_bucket_size': hidden_size * hidden_size,
                        'zero_optimization.stage3_param_persistence_threshold': 10 * hidden_size,
                        'zero_optimization.stage3_prefetch_bucket_size': 0.9 * hidden_size * hidden_size,
                    }
                )

    if config_kwargs['zero_optimization']['stage'] != 3:
        config_kwargs['zero_optimization']['stage'] = 0
    config_kwargs['optimizer'] = {'type': None}

    model, *_ = deepspeed.initialize(model=model, config=config_kwargs)
    model.eval()
    return model


class DPOLossRegistry(Registrable):
    @abstractmethod
    def compute_loss(
        self,
        policy_chosen_logps: torch.FloatTensor,
        policy_rejected_logps: torch.FloatTensor,
        reference_chosen_logps: torch.FloatTensor,
        reference_rejected_logps: torch.FloatTensor,
        precomputed_margins: torch.FloatTensor | None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        ...



from typing import Literal

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.cherry_pick import ClassificationCherryPickSettings
from turbo_alignment.settings.datasets.classification import (
    ClassificationMultiDatasetSettings,
)
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings
from turbo_alignment.settings.tf.trainer import TrainerSettings


class ClassificationLossSettings(ExtraFieldsNotAllowedBaseModel):
    alpha: Literal['auto'] | list[float] | None
    gamma: float


class ClassificationTrainerSettings(TrainerSettings):
    loss_settings: ClassificationLossSettings


class ClassificationTrainExperimentSettings(BaseTrainExperimentSettings):
    trainer_settings: ClassificationTrainerSettings

    train_dataset_settings: ClassificationMultiDatasetSettings
    val_dataset_settings: ClassificationMultiDatasetSettings

    cherry_pick_settings: ClassificationCherryPickSettings



from abc import ABC

import torch

from turbo_alignment.modeling.common.helpers import SelectElement, SelectEOSAndProject
from turbo_alignment.modeling.imagebind.models import (
    ImageBindArchitectureSettings,
    ModalityType,
)


class Heads(ABC, torch.nn.ModuleDict):
    @staticmethod
    def __get_vision_head(
        vision_embed_dim: int,
        out_embed_dim: int,
    ) -> torch.nn.Sequential:
        return torch.nn.Sequential(
            torch.nn.LayerNorm(normalized_shape=vision_embed_dim, eps=1e-6),
            SelectElement(index=0),
            torch.nn.Linear(vision_embed_dim, out_embed_dim, bias=False),
        )

    @staticmethod
    def __get_text_head(
        text_embed_dim: int,
        out_embed_dim: int,
    ) -> SelectEOSAndProject:
        return SelectEOSAndProject(
            proj=torch.nn.Sequential(
                torch.nn.LayerNorm(normalized_shape=text_embed_dim, eps=1e-6),
                torch.nn.Linear(text_embed_dim, out_embed_dim, bias=False),
            )
        )

    @staticmethod
    def __get_audio_head(
        audio_embed_dim: int,
        out_embed_dim: int,
    ) -> torch.nn.Sequential:
        return torch.nn.Sequential(
            torch.nn.LayerNorm(normalized_shape=audio_embed_dim, eps=1e-6),
            SelectElement(index=0),
            torch.nn.Linear(audio_embed_dim, out_embed_dim, bias=False),
        )

    def __init__(self, settings: ImageBindArchitectureSettings):
        super().__init__()

        self[ModalityType.VISION] = self.__get_vision_head(settings.vision_embed_dim, settings.out_embed_dim)
        self[ModalityType.TEXT] = self.__get_text_head(settings.text_embed_dim, settings.out_embed_dim)
        self[ModalityType.AUDIO] = self.__get_audio_head(settings.audio_embed_dim, settings.out_embed_dim)



from typing import Literal

from turbo_alignment.settings.datasets.base import (
    BaseDatasetSettings,
    DatasetType,
    MultiDatasetSettings,
)
from turbo_alignment.settings.datasets.chat import ChatDatasetSettings


class ClassificationDatasetSettings(BaseDatasetSettings):
    dataset_type: Literal[DatasetType.CLASSIFICATION] = DatasetType.CLASSIFICATION
    chat_settings: ChatDatasetSettings


class ClassificationMultiDatasetSettings(ClassificationDatasetSettings, MultiDatasetSettings):
    ...



from collections import defaultdict
from typing import Any, Callable

import pandas as pd
from accelerate.utils.operations import gather_object
from transformers import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)
from transformers.trainer_callback import CallbackHandler

from turbo_alignment.settings.metric import MetricResults


class EvaluateFirstStepCallback(TrainerCallback):
    def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        if state.global_step == 0:
            control.should_evaluate = True

        return control


class MetricsCallbackHandler(CallbackHandler):
    def __init__(self, *args, **kwargs) -> None:
        self.ref_model = kwargs.pop('ref_model', None)
        self.sft_model = kwargs.pop('sft_model', None)
        self.accelerator = kwargs.pop('accelerator', None)
        self.metrics_kwargs = kwargs.pop('metrics_kwargs', {})
        super().__init__(*args, **kwargs)

    def on_evaluate(
        self, args: TrainingArguments, state: TrainerState, control: TrainerControl, metrics
    ) -> TrainerControl:
        control.should_evaluate = False

        results = self.call_event(
            'on_evaluate',
            args,
            state,
            control,
            metrics=metrics,
            ref_model=self.ref_model,
            sft_model=self.sft_model,
            accelerator=self.accelerator,
            metrics_kwargs=self.metrics_kwargs,
        )
        if isinstance(results, list):
            gathered_results: list[list[MetricResults]] = gather_object(results)

            gathered_float_data: dict[str, list[Any]] = defaultdict(list)
            gathered_table_data: dict[str, list[str] | list[list[str]]] = defaultdict(list)
            average_functions: dict[str, Callable] = {}

            for single_process_results in gathered_results:
                for metric_result in single_process_results:
                    for score in metric_result.element_wise_scores:
                        if metric_result.need_average:
                            gathered_float_data[score.label].extend(score.values)
                            average_functions[score.label] = score.average_function
                        else:
                            gathered_table_data[score.label].extend(score.values)

            logs = {
                'cherry_pick_' + k: average_functions[k](list(*zip(*v)))
                if isinstance(v[0], tuple)
                else average_functions[k](v)
                for k, v in gathered_float_data.items()
            }

            self.call_event('on_log', args, state, control, logs=logs)

            table_cols = list(gathered_table_data.keys())
            table_data = list(gathered_table_data.values())

            flattened_table_data = [
                sum(item, []) if isinstance(item, list) and isinstance(item[0], list) else item for item in table_data
            ]  # flatten list[lists] to display all outputs in wandb table

            cherrypicks_table_data = pd.DataFrame(columns=table_cols, data=list(zip(*flattened_table_data)))
            dataset_prefixes = set(col.split('@@')[0] for col in cherrypicks_table_data.columns)

            for dataset_prefix in dataset_prefixes:
                dataset_columns = [col for col in cherrypicks_table_data.columns if col.startswith(dataset_prefix)]

                table = {
                    f'cherry_pick_table_{dataset_prefix}_{state.global_step}': cherrypicks_table_data[dataset_columns]
                }

                self.call_event('on_log', args, state, control, logs=table)

        return control



from typing import Any

import torch
from transformers import PreTrainedTokenizerBase
from vllm import LLM, SamplingParams
from vllm.lora.request import LoRARequest

from turbo_alignment.dataset.chat import ChatDatasetRecord
from turbo_alignment.generators.base import BaseGenerator
from turbo_alignment.settings.generators.chat import CustomChatGenerationSettings
from turbo_alignment.settings.generators.outputs.chat import (
    AnswerMessage,
    ChatInferenceOutput,
)
from turbo_alignment.settings.tf.generation import GeneratorTransformersSettings


class VLLMChatGenerator(BaseGenerator[ChatDatasetRecord, ChatInferenceOutput]):
    def __init__(
        self,
        transformers_settings: GeneratorTransformersSettings,
        custom_generation_settings: CustomChatGenerationSettings,
        model: LLM,
        tokenizer: PreTrainedTokenizerBase,
        batch: int,
        return_logits: bool = False,
        lora_request: LoRARequest | None = None,
    ):
        model.set_tokenizer(tokenizer)
        super().__init__(model, tokenizer, batch=batch)

        if isinstance(transformers_settings.stop_strings, list):
            raise ValueError('You should use only 1 eos token with VLLM')

        eos_token_id: list[int] = self._tokenizer.encode(transformers_settings.stop_strings, add_special_tokens=False)

        beam_search_params: dict[str, Any] = {
            'best_of': transformers_settings.num_return_sequences,
            'use_beam_search': False,
        }
        if transformers_settings.num_beams > 1:
            beam_search_params['use_beam_search'] = True
            beam_search_params['best_of'] = transformers_settings.num_beams

        self._sampling_params = SamplingParams(
            n=transformers_settings.num_return_sequences,
            repetition_penalty=transformers_settings.repetition_penalty,
            temperature=transformers_settings.temperature,
            top_p=transformers_settings.top_p,
            top_k=transformers_settings.top_k,
            skip_special_tokens=custom_generation_settings.skip_special_tokens,
            stop_token_ids=eos_token_id,
            max_tokens=transformers_settings.max_new_tokens,
            **beam_search_params,
        )
        self._lora_request = lora_request

        self._return_logits = return_logits

    def _generate_from_batch(
        self, records: list[dict[str, Any]], original_records: list[ChatDatasetRecord], dataset_name: str
    ) -> list[ChatInferenceOutput]:
        input_ids = [record['input_ids'].tolist() for record in records]
        request_outputs = self._model.generate(
            prompts=None,
            prompt_token_ids=input_ids,
            sampling_params=self._sampling_params,
            lora_request=self._lora_request,
        )

        outputs = []
        for i, request_output in enumerate(request_outputs):
            original_record = original_records[i]
            answers = []
            for a in request_output.outputs:
                ans_msg = AnswerMessage(
                    id=str(a.index),
                    content=a.text,
                    sequence_score=a.cumulative_logprob,
                )
                if self._return_logits:
                    ans_msg.input_token_ids = torch.tensor(request_output.prompt_token_ids).unsqueeze(0)
                    ans_msg.answer_token_ids = torch.tensor(a.token_ids).unsqueeze(0)

                answers.append(ans_msg)

            outputs.append(
                ChatInferenceOutput(
                    id=original_record.id,
                    dataset_name=dataset_name,
                    messages=original_record.messages,
                    label=original_record.label,
                    answers=answers,
                )
            )
        return outputs



import torch

from turbo_alignment.dataset.chat import ChatDatasetRecord
from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.generators.outputs.base import BaseInferenceOutput


class AnswerMessage(ExtraFieldsNotAllowedBaseModel):
    id: str
    content: str
    sequence_score: float | None = None
    input_token_ids: torch.Tensor | None = None
    answer_token_ids: torch.Tensor | None = None
    logits: torch.Tensor | None = None

    class Config:
        arbitrary_types_allowed = True


class ChatInferenceOutput(BaseInferenceOutput, ChatDatasetRecord):
    answers: list[AnswerMessage]


class RagInferenceOutput(ChatInferenceOutput):
    documents: list[str]
    doc_scores: list[float]



from enum import Enum

from turbo_alignment.common.registry import Registrable
from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.datasets.chat import ChatPromptTemplate
from turbo_alignment.settings.metric import MetricType
from turbo_alignment.settings.model import (
    PreTrainedAdaptersModelSettings,
    PreTrainedModelSettings,
)
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings


class KLType(str, Enum):
    SFT_MODEL: str = 'sft'
    REFERENCE_MODEL: str = 'reference'


class MetricSettingsRegistry(Registrable):
    ...


class MetricSettings(ExtraFieldsNotAllowedBaseModel):
    need_average: list[bool]


@MetricSettingsRegistry.register(MetricType.DIST_N)
class DistinctnessSettings(MetricSettings):
    ...


@MetricSettingsRegistry.register(MetricType.DIVERSITY)
class DiversitySettings(MetricSettings):
    ...


@MetricSettingsRegistry.register(MetricType.LENGTH)
class LengthSettings(MetricSettings):
    ...


@MetricSettingsRegistry.register(MetricType.KL)
class KLSettings(MetricSettings):
    ref_logits_type: KLType


@MetricSettingsRegistry.register(MetricType.METEOR)
class MeteorSettings(MetricSettings):
    compute_element_wise_meteor: bool = True
    element_wise_meteor_label: str = 'meteor'


@MetricSettingsRegistry.register(MetricType.PERPLEXITY)
class PerplexitySettings(MetricSettings):
    ...


@MetricSettingsRegistry.register(MetricType.REWARD)
class RewardSettings(MetricSettings):
    tokenizer_settings: TokenizerSettings
    model_settings: PreTrainedModelSettings | PreTrainedAdaptersModelSettings
    prompt_template: ChatPromptTemplate
    max_tokens_count: int
    system_prompt: str | None = None
    name: str = 'cherry_pick_dataset'
    micro_batch: int = 1


@MetricSettingsRegistry.register(MetricType.ROUGE)
class RougeSettings(MetricSettings):
    need_compute_rouge_n: list[int]
    need_compute_rouge_l: bool = True


@MetricSettingsRegistry.register(MetricType.SELF_BLEU)
class SelfBleuSettings(MetricSettings):
    ngram: int = 3


@MetricSettingsRegistry.register(MetricType.TOOL_CALL_METRICS)
class ToolMetricsSettings(MetricSettings):
    tool_activation_str: str


@MetricSettingsRegistry.register(MetricType.RETRIEVAL_UTILITY)
class RetrievalUtilitySettings(MetricSettings):
    doc_sep_symbol: str = '<doc_sep>'



import os
import pathlib
from pathlib import Path

import boto3

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.settings.s3 import S3HandlerParameters

logger = get_project_logger()


class S3CheckpointsHandler:
    def __init__(self, parameters: S3HandlerParameters) -> None:
        self._resource = boto3.resource(
            service_name='s3',
            endpoint_url=parameters.host,
            aws_access_key_id=parameters.aws_access_key_id,
            aws_secret_access_key=parameters.aws_secret_access_key,
        )

        self._parameters = parameters

    def upload_local_files(self, path: pathlib.Path, target_directory: str, save_optimizer_states: bool) -> None:
        file_names_count = len(list(os.walk(path)))

        for root, _, files in os.walk(path):
            for index, file in enumerate(files):
                if not save_optimizer_states and 'global_step' in root:
                    continue

                root_path = Path(root)
                file_path = root_path / file

                target_directory_path = Path(target_directory)

                key = target_directory_path / file_path.relative_to(path)

                logger.info(
                    f'''🥭 Start to upload file {index} / {file_names_count}
                                    from local {file} to remote s3 with key {key}'''
                )

                self._resource.meta.client.upload_file(
                    Filename=str(file_path), Bucket=self._parameters.bucket, Key=str(key)
                )

                logger.info(
                    f'''🥝 Successfully uploaded file {index} / {file_names_count}
                                    from local {file_path} to remote s3 with key {key}'''
                )

    def upload_local_file(self, file_path: Path, target_directory: Path) -> None:
        self._resource.meta.client.upload_file(
            Filename=str(file_path), Bucket=self._parameters.bucket, Key=str(target_directory / file_path.name)
        )

    def directory_exists(self, directory_path: Path) -> bool:
        objects = self._resource.Bucket(self._parameters.bucket).objects.filter(Prefix=f'{directory_path}/')
        return len([o.key for o in objects]) > 0

    def remove_directory(self, prefix: str) -> None:
        logger.info(f'🥭 Start to remove directory with prefix {prefix}')
        bucket = self._resource.Bucket(self._parameters.bucket)
        bucket.objects.filter(Prefix=prefix).delete()
        logger.info(f'🥝 Successfully removed directory with prefix {prefix}')

    def list_folders_keys(self, folder: str) -> list[str]:
        folders = []

        result = self._resource.meta.client.list_objects(
            Bucket=self._parameters.bucket,
            Prefix=f'{folder}/',
            Delimiter='/',
            MaxKeys=1_000,
        )

        prefixes = result.get('CommonPrefixes', [])
        for prefix in prefixes:
            _folder = prefix['Prefix'].split('/')
            # folder always ends in / so -1 will
            # point to empty string
            right_most_folder_name = _folder[-2]
            folders.append(right_most_folder_name)

        return folders



from turbo_alignment.modeling.multimodal.encoders.audio.imagebind import (
    ImageBindAudioModeling,
)
from turbo_alignment.modeling.multimodal.encoders.audio.whisper import (
    WhisperAudioModeling,
)
from turbo_alignment.modeling.multimodal.encoders.base import BaseModalityEncoder
from turbo_alignment.modeling.multimodal.encoders.image.clip import CLIPImageModeling
from turbo_alignment.modeling.multimodal.encoders.image.imagebind import (
    ImageBindImageModeling,
)
from turbo_alignment.modeling.multimodal.encoders.registry import (
    ModalityEncoderRegistry,
)



from pathlib import Path

from pydantic import field_validator
from pydantic_settings import BaseSettings


class S3HandlerParametersWrongBucketException(Exception):
    ...


class S3HandlerParameters(BaseSettings):
    host: str
    bucket: str
    aws_access_key_id: str
    aws_secret_access_key: str

    @field_validator('bucket')
    def bucket_name_biglm_is_not_allowed(cls, values: str) -> str:
        if values == 'biglm':
            raise S3HandlerParametersWrongBucketException('Usage of biglm bucket is not allowed')

        return values

    class Config:
        env_file: str = '.env'
        env_prefix: str = 'S3_CHECKPOINTS_'


class CheckpointUploaderCallbackParameters(BaseSettings):
    directory: Path

    save_optimizer_states: bool = True
    tokenizer_folder_name: str = 'tokenizer'
    checkpoints_folder_name: str = 'checkpoints'

    @property
    def tokenizer_directory(self) -> str:
        return f'{self.directory}/{self.tokenizer_folder_name}'

    @property
    def checkpoints_directory(self) -> str:
        return f'{self.directory}/{self.checkpoints_folder_name}'

    class Config:
        env_prefix: str = 'CHECKPOINT_UPLOADER_CALLBACK_'


class ExperimentMetadata(BaseSettings):
    wandb_run_id: str = 'wandb_run_id'
    platform_job_name: str = 'platform_job_name'
    last_git_commit_hash: str = 'last_git_commit_hash'



import math
from abc import ABC
from typing import Callable, Literal

import numpy as np
import torch
from timm.models.layers import trunc_normal_

from turbo_alignment.modeling.common.helpers import VerboseNNModule, cast_if_src_dtype


def _get_sinusoid_encoding_table(n_position: int, d_hid: int) -> torch.Tensor:
    """Sinusoid position encoding table"""

    def get_position_angle_vec(position):
        return [position / np.power(10000, 2 * (hid_j // 2) / d_hid) for hid_j in range(d_hid)]

    sinusoid_table = np.array([get_position_angle_vec(pos_i) for pos_i in range(n_position)])
    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1

    return torch.FloatTensor(sinusoid_table).unsqueeze(0)


def _interpolate_pos_encoding_2d(target_spatial_size: int, pos_embed: torch.Tensor) -> torch.Tensor:
    N = pos_embed.shape[1]
    if N == target_spatial_size:
        return pos_embed
    dim = pos_embed.shape[-1]
    # torch.nn.functional.interpolate doesn't work with bfloat16, so we cast to float32
    pos_embed, updated = cast_if_src_dtype(pos_embed, torch.bfloat16, torch.float32)
    pos_embed = torch.nn.functional.interpolate(
        pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
        scale_factor=math.sqrt(target_spatial_size / N),
        mode='bicubic',
    )
    if updated:
        pos_embed, _ = cast_if_src_dtype(pos_embed, torch.float32, torch.bfloat16)
    pos_embed = pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
    return pos_embed


def _interpolate_pos_encoding(
    npatch_per_img: int,
    pos_embed: torch.Tensor,
    patches_layout: list,
    input_shape: tuple | None = None,
    first_patch_idx: int = 1,
):
    assert first_patch_idx in (0, 1), 'there is 1 CLS token or none'
    N = pos_embed.shape[1] - first_patch_idx  # since it's 1 if cls_token exists
    if npatch_per_img == N:
        return pos_embed

    assert patches_layout[-1] == patches_layout[-2], 'Interpolation of pos embed not supported for non-square layouts'

    class_emb = pos_embed[:, :first_patch_idx]
    pos_embed = pos_embed[:, first_patch_idx:]

    if input_shape is None or patches_layout[0] == 1:
        # simple 2D pos embedding, no temporal component
        pos_embed = _interpolate_pos_encoding_2d(npatch_per_img, pos_embed)
    elif patches_layout[0] > 1:
        # pos embed has a temporal component
        assert len(input_shape) == 4, 'temporal interpolation not supported'
        # we only support 2D interpolation in this case
        num_frames = patches_layout[0]
        num_spatial_tokens = patches_layout[1] * patches_layout[2]
        pos_embed = pos_embed.view(1, num_frames, num_spatial_tokens, -1)
        # interpolate embedding for zeroth frame
        pos_embed = _interpolate_pos_encoding_2d(npatch_per_img, pos_embed[0, 0, ...].unsqueeze(0))
    else:
        raise ValueError("This type of interpolation isn't implemented")

    return torch.cat((class_emb, pos_embed), dim=1)


def _build_causal_attention_mask(context_length: int) -> torch.Tensor:
    # lazily create causal attention mask, with full attention between the vision tokens
    # pytorch uses additive attention mask; fill with -inf
    mask = torch.empty(context_length, context_length, requires_grad=False)
    mask.fill_(float('-inf'))
    mask.triu_(1)  # zero out the lower diagonal
    return mask


def _get_pos_embedding(
    npatch_per_img: int,
    pos_embed: torch.Tensor,
    patches_layout: list,
    input_shape: tuple,
    first_patch_idx=1,
):
    pos_embed = _interpolate_pos_encoding(
        npatch_per_img,
        pos_embed,
        patches_layout,
        input_shape=input_shape,
        first_patch_idx=first_patch_idx,
    )
    return pos_embed


class PatchEmbedGeneric(torch.nn.Module):
    """
    PatchEmbed from Hydra
    """

    def __init__(self, proj_stem: list[torch.nn.Module], norm_layer: torch.nn.Module | None = None):
        super().__init__()

        self.proj: torch.nn.Module
        if len(proj_stem) > 1:
            self.proj = torch.nn.Sequential(*proj_stem)
        else:
            # Special case to be able to load pre-trained models that were
            # trained with a standard stem
            self.proj = proj_stem[0]

        self.norm_layer = norm_layer

    def get_patch_layout(self, img_size):
        with torch.no_grad():
            dummy_img = torch.zeros(
                [
                    1,
                ]
                + img_size
            )
            dummy_out = self.proj(dummy_img)
        embed_dim = dummy_out.shape[1]
        patches_layout = tuple(dummy_out.shape[2:])
        num_patches = np.prod(patches_layout)
        return patches_layout, num_patches, embed_dim

    def forward(self, x):
        with torch.no_grad():
            x = self.proj(x)
            # B C (T) H W -> B (T)HW C
            x = x.flatten(2).transpose(1, 2)
            if self.norm_layer is not None:
                x = self.norm_layer(x)
        return x


class SpatioTemporalPosEmbeddingHelper(VerboseNNModule, ABC):
    def __init__(
        self,
        patches_layout: list,
        num_patches: int,
        num_cls_tokens: int,
        embed_dim: int,
        learnable: bool,
    ) -> None:
        super().__init__()
        self.num_cls_tokens = num_cls_tokens
        self.patches_layout = patches_layout
        self.num_patches = num_patches
        self.num_tokens = num_cls_tokens + num_patches
        self.learnable = learnable
        if self.learnable:
            self.pos_embed = torch.nn.Parameter(torch.zeros(1, self.num_tokens, embed_dim))
            trunc_normal_(self.pos_embed, std=0.02)
        else:
            self.register_buffer('pos_embed', _get_sinusoid_encoding_table(self.num_tokens, embed_dim))

    def get_pos_embedding(self, vision_input, all_vision_tokens):
        input_shape = vision_input.shape
        pos_embed = _get_pos_embedding(
            all_vision_tokens.size(1) - self.num_cls_tokens,
            pos_embed=self.pos_embed,
            patches_layout=self.patches_layout,
            input_shape=input_shape,
            first_patch_idx=self.num_cls_tokens,
        )
        return pos_embed


class RGBDTPreprocessor(VerboseNNModule):
    def __init__(
        self,
        rgbt_stem: PatchEmbedGeneric | None,
        depth_stem: PatchEmbedGeneric | None,
        img_size: list | None = None,
        num_cls_tokens: int = 1,
        pos_embed_fn: Callable | None = None,
        use_type_embed: bool = False,
        init_param_style: str = 'openclip',
    ) -> None:
        super().__init__()
        stem = rgbt_stem if rgbt_stem is not None else depth_stem
        if img_size is None:
            img_size = [3, 224, 224]
        assert stem is not None
        (
            self.patches_layout,
            self.num_patches,
            self.embed_dim,
        ) = stem.get_patch_layout(img_size)
        self.rgbt_stem = rgbt_stem
        self.depth_stem = depth_stem
        self.use_pos_embed = pos_embed_fn is not None
        self.use_type_embed = use_type_embed
        self.num_cls_tokens = num_cls_tokens

        if pos_embed_fn is not None:
            self.pos_embedding_helper = pos_embed_fn(
                patches_layout=self.patches_layout,
                num_cls_tokens=num_cls_tokens,
                num_patches=self.num_patches,
                embed_dim=self.embed_dim,
            )
        if self.num_cls_tokens > 0:
            self.cls_token = torch.nn.Parameter(torch.zeros(1, self.num_cls_tokens, self.embed_dim))
        if self.use_type_embed:
            self.type_embed = torch.nn.Parameter(torch.zeros(1, 1, self.embed_dim))

        self.init_parameters(init_param_style)

    @torch.no_grad()
    def init_parameters(self, init_param_style):
        if init_param_style == 'openclip':
            # OpenCLIP style initialization
            scale = self.embed_dim**-0.5
            if self.use_pos_embed:
                torch.nn.init.normal_(self.pos_embedding_helper.pos_embed)
                self.pos_embedding_helper.pos_embed *= scale

            if self.num_cls_tokens > 0:
                torch.nn.init.normal_(self.cls_token)
                self.cls_token *= scale
        elif init_param_style == 'vit':
            self.cls_token.data.fill_(0)
        else:
            raise ValueError(f'Unknown init {init_param_style}')

        if self.use_type_embed:
            torch.nn.init.normal_(self.type_embed)

    def tokenize_input_and_cls_pos(self, inputs, stem):
        # tokens is of shape B x L x D
        tokens = stem(inputs)
        assert tokens.ndim == 3
        assert tokens.shape[2] == self.embed_dim
        B = tokens.shape[0]
        if self.num_cls_tokens > 0:
            class_tokens = self.cls_token.expand(B, -1, -1)  # stole class_tokens impl from Phil Wang, thanks
            tokens = torch.cat((class_tokens, tokens), dim=1)
        if self.use_pos_embed:
            pos_embed = self.pos_embedding_helper.get_pos_embedding(inputs, tokens)
            tokens = tokens + pos_embed
        if self.use_type_embed:
            tokens = tokens + self.type_embed.expand(B, -1, -1)
        return tokens

    def forward(self, vision: torch.Tensor | None = None, depth: torch.Tensor | None = None, patch_mask=None) -> dict:
        if patch_mask is not None:
            raise NotImplementedError()

        if vision is not None:
            vision_tokens = self.tokenize_input_and_cls_pos(vision, self.rgbt_stem)

        if depth is not None:
            depth_tokens = self.tokenize_input_and_cls_pos(depth, self.depth_stem)

        # aggregate tokens
        if vision is not None and depth is not None:
            final_tokens = vision_tokens + depth_tokens
        else:
            final_tokens = vision_tokens if vision is not None else depth_tokens
        return_dict = {
            'trunk': {
                'tokens': final_tokens,
            },
            'head': {},
        }
        return return_dict


class AudioPreprocessor(RGBDTPreprocessor):
    def __init__(self, audio_stem: PatchEmbedGeneric, **kwargs) -> None:
        super().__init__(rgbt_stem=audio_stem, depth_stem=None, **kwargs)

    def forward(self, *_args, audio: torch.Tensor | None = None, **_kwargs) -> dict:
        # vision here is actually audio
        return super().forward(vision=audio)


class TextPreprocessor(VerboseNNModule):
    def __init__(
        self,
        vocab_size: int,
        context_length: int,
        embed_dim: int,
        causal_masking: bool,
        supply_seq_len_to_head: bool = True,
        num_cls_tokens: int = 0,
        init_param_style: str = 'openclip',
    ) -> None:
        super().__init__()
        self.vocab_size = vocab_size
        self.context_length = context_length
        self.token_embedding = torch.nn.Embedding(vocab_size, embed_dim)
        self.pos_embed = torch.nn.Parameter(torch.empty(1, self.context_length + num_cls_tokens, embed_dim))
        self.causal_masking = causal_masking
        if self.causal_masking:
            mask = _build_causal_attention_mask(self.context_length)
            # register the mask as a buffer so it can be moved to the right device
            self.register_buffer('mask', mask)

        self.supply_seq_len_to_head = supply_seq_len_to_head
        self.num_cls_tokens = num_cls_tokens
        self.embed_dim = embed_dim
        if num_cls_tokens > 0:
            assert self.causal_masking is False, "Masking + CLS token isn't implemented"
            self.cls_token = torch.nn.Parameter(torch.zeros(1, self.num_cls_tokens, embed_dim))

        self.init_parameters(init_param_style)

    @torch.no_grad()
    def init_parameters(self, init_param_style='openclip'):
        # OpenCLIP style initialization
        torch.nn.init.normal_(self.token_embedding.weight, std=0.02)
        torch.nn.init.normal_(self.pos_embed, std=0.01)

        if init_param_style == 'openclip':
            # OpenCLIP style initialization
            scale = self.embed_dim**-0.5
            if self.num_cls_tokens > 0:
                torch.nn.init.normal_(self.cls_token)
                self.cls_token *= scale
        elif init_param_style == 'vit':
            self.cls_token.data.fill_(0)
        else:
            raise ValueError(f'Unknown init {init_param_style}')

    def forward(self, text: torch.Tensor) -> dict:
        # text tokens are of shape B x L x D
        text_tokens = self.token_embedding(text)
        # concat CLS tokens if any
        if self.num_cls_tokens > 0:
            B = text_tokens.shape[0]
            class_tokens = self.cls_token.expand(B, -1, -1)  # stole class_tokens impl from Phil Wang, thanks
            text_tokens = torch.cat((class_tokens, text_tokens), dim=1)
        text_tokens = text_tokens + self.pos_embed
        return_dict: dict[str, dict] = {
            'trunk': {
                'tokens': text_tokens,
            },
            'head': {},
        }
        # Compute sequence length after adding CLS tokens
        if self.supply_seq_len_to_head:
            text_lengths = text.argmax(dim=-1)
            return_dict['head'] = {
                'seq_len': text_lengths,
            }
        if self.causal_masking:
            return_dict['trunk'].update({'attn_mask': self.mask})
        return return_dict


class Im2Video(torch.nn.Module):
    """Convert an image into a trivial video."""

    def __init__(self, time_dim: int = 2):
        super().__init__()
        self.time_dim = time_dim

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if x.ndim == 4:
            # B, C, H, W -> B, C, T, H, W
            return x.unsqueeze(self.time_dim)
        if x.ndim == 5:
            return x
        raise ValueError(f'Dimension incorrect {x.shape}')


class PadIm2Video(Im2Video):
    def __init__(self, ntimes: int, pad_type: Literal['zero', 'repeat'], time_dim: int = 2):
        super().__init__(time_dim=time_dim)
        assert ntimes > 0
        self.ntimes = ntimes
        self.pad_type = pad_type

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = super().forward(x)
        if x.shape[self.time_dim] == 1:
            if self.pad_type == 'repeat':
                new_shape = [1] * len(x.shape)
                new_shape[self.time_dim] = self.ntimes
                x = x.repeat(new_shape)
            elif self.pad_type == 'zero':
                padarg = [0, 0] * len(x.shape)
                padarg[2 * self.time_dim + 1] = self.ntimes - x.shape[self.time_dim]
                x = torch.nn.functional.pad(x, padarg)
        return x



from collections import defaultdict
from pathlib import Path

import torch
from scipy.ndimage.measurements import find_objects, label
from transformers.utils import ModelOutput

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.modeling.multimodal.lm.base import BaseMultiModalModeling
from turbo_alignment.modeling.multimodal.projectors import LlavaMultiModalProjector
from turbo_alignment.modeling.multimodal.projectors.registry import (
    MultiModalProjectorRegistry,
)
from turbo_alignment.settings.modality import Modality

logger = get_project_logger()


class ProjectionMultiModalModeling(BaseMultiModalModeling):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        modality_adapters = {}

        for modality in self.encoders:
            modality_adapters[modality] = (
                MultiModalProjectorRegistry.by_name(self.modality_projector_mapping[modality])(
                    encoder_hidden_size=self.encoders[modality].emb_dim,
                    text_hidden_size=self.language_model_dim,
                    n_modality_embs=self.n_modality_embs,
                )
                .to(self.language_model.device)
                .to(torch.bfloat16)
            )

            assert (
                not isinstance(modality_adapters[modality], LlavaMultiModalProjector)
                or self.n_modality_embs == self.encoders[modality].n_modality_embs
            ), f'For the LLaVA MLP projector, n_modality_embs should be equal to the \
total number of patches returned by the encoder. \
Please, set n_modality_embs to {self.encoders[modality].n_modality_embs} in config.'

            if self.modality_projector_initialization_mapping:
                state_dict_path: Path | None = self.modality_projector_initialization_mapping.get(modality, None)
                if state_dict_path is not None:
                    logger.info(f'Loading {modality} connector weights')

                    state_dictionary = torch.load(state_dict_path)
                    modality_adapters[modality].load_state_dict(state_dictionary)
                    logger.info(f'Sucsessfully loaded from {state_dict_path}')

        self.modality_adapters = torch.nn.ModuleDict(modality_adapters)

        self.config = self.language_model.config

    def convert_inputs_to_embeds(
        self,
        input_ids: torch.LongTensor,
        modality_inputs: list[list[tuple[Modality, torch.Tensor]]],
        modality_tokens_mask: torch.Tensor,
    ):
        multimodal_lm_input_embeds: list[torch.Tensor] = []

        lm_input_embeds = self.language_model.base_model.model.model.embed_tokens.modules_to_save.default(input_ids)

        for sample_lm_input_embeds, sample_modality_tokens_mask, sample_modality_inputs in zip(
            lm_input_embeds, modality_tokens_mask, modality_inputs
        ):
            span_mask, _ = label(
                sample_modality_tokens_mask.cpu().detach().numpy()
            )  # returns mask with ids of spans from 1 to N
            modality_spans = find_objects(span_mask)  # returns list of tuples with start index and end index

            assert len(modality_spans) == len(sample_modality_inputs)

            grouped_modality_encoder_inputs: dict[Modality, list[tuple[int, torch.Tensor]]] = defaultdict(list)

            # Prepare modality batches
            for index, modality_object in enumerate(sample_modality_inputs):
                modality, inputs = modality_object
                grouped_modality_encoder_inputs[modality].append((index, inputs))

            sorted_modality_embeddings: torch.Tensor = torch.full(
                (len(sample_modality_inputs), self.n_modality_embs, self.language_model_dim), torch.nan
            ).to(self.language_model.device)

            # Encode modalities and insert into input embeds
            for modality, modality_encoder_inputs_with_indices in grouped_modality_encoder_inputs.items():
                modality_encoder_input_indexes, modality_encoder_inputs = zip(*modality_encoder_inputs_with_indices)

                if self.language_model.dtype == torch.float32:
                    encoded_modality_object_batch = self.encoders[modality].encode(
                        torch.stack(modality_encoder_inputs, dim=0).to(self.language_model.device)
                    )
                else:
                    encoded_modality_object_batch = self.encoders[modality].encode(
                        torch.stack(modality_encoder_inputs, dim=0).to(self.language_model.device).bfloat16()
                    )

                modality_encoder_embeddings = self.modality_adapters[modality](encoded_modality_object_batch)

                sorted_modality_embeddings[modality_encoder_input_indexes, :] = modality_encoder_embeddings.to(
                    sorted_modality_embeddings.dtype
                )

            substituted_sample_lm_input_embeds = sample_lm_input_embeds.clone()
            for i, modality_span in enumerate(modality_spans):
                substituted_sample_lm_input_embeds[
                    modality_span[0].start : modality_span[0].stop
                ] = sorted_modality_embeddings[i, :]

            multimodal_lm_input_embeds.append(substituted_sample_lm_input_embeds)

        return torch.stack(multimodal_lm_input_embeds)

    def forward(
        self,
        input_ids: torch.LongTensor,
        modality_inputs: list[list[tuple[Modality, torch.Tensor]]],
        attention_mask: torch.LongTensor,
        modality_tokens_mask: torch.Tensor,
        labels: torch.LongTensor | None = None,
    ) -> ModelOutput:
        multimodal_lm_input_embeds = self.convert_inputs_to_embeds(input_ids, modality_inputs, modality_tokens_mask)
        return self.language_model(
            inputs_embeds=multimodal_lm_input_embeds, labels=labels, attention_mask=attention_mask
        )



from abc import ABC, abstractmethod

import torch


class BaseModalityEncoder(torch.nn.Module, ABC):
    @abstractmethod
    def encode(self, inputs: torch.Tensor) -> torch.Tensor:
        ...

    @property
    @abstractmethod
    def emb_dim(self) -> int:
        ...

    @property
    @abstractmethod
    def device(self) -> int:
        ...

    @property
    @abstractmethod
    def n_modality_embs(self) -> int:
        ...



from pathlib import Path

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.model import (
    ModelForPeftSettings,
    PreTrainedAdaptersModelSettings,
    PreTrainedModelSettings,
)


class RAGIndexSettings(ExtraFieldsNotAllowedBaseModel):
    index_path: Path
    passages_path: Path


class RAGRetrievalSettings(ExtraFieldsNotAllowedBaseModel):
    n_docs: int = 1
    title_sep: str | None = ' / '
    doc_sep: str | None = ' // '
    prefix: str | None = ''  # 'query:' for e5
    max_doc_length: int = 300
    retrieval_vector_size: int = 768
    retrieval_batch_size: int = 8
    query_encoder_max_length: int = 512
    encoder_learning_rate: float | None = None


class RAGPreTrainedModelSettings(ExtraFieldsNotAllowedBaseModel):
    generator_settings: ModelForPeftSettings | PreTrainedAdaptersModelSettings
    question_encoder_settings: PreTrainedModelSettings
    index_settings: RAGIndexSettings
    retrieval_settings: RAGRetrievalSettings
